"""
Unit tests for the
"""

import json
import treq

from twisted.trial.unittest import SynchronousTestCase
from mimic.canned_responses.loadbalancer import load_balancer_example
from mimic.model.clb_errors import (
    considered_immutable_error,
    invalid_json_schema,
    loadbalancer_not_found,
    node_not_found,
    updating_node_validation_error
)
from mimic.test.fixtures import APIMockHelper, TenantAuthentication
from mimic.rest.loadbalancer_api import LoadBalancerApi, LoadBalancerControlApi
from mimic.test.helpers import json_request, request_with_content, request
from mimic.util.helper import EMPTY_RESPONSE
import attr


class ResponseGenerationTests(SynchronousTestCase):
    """
    Tests for Loud Balancer response generation.
    """

    def test_canned_loadbalancer(self):
        """
        Test that the canned load balancer response is returned as expected.
        """

        expect_lb_name = "cannedTestLB"
        expect_lb_protocol = "protocol"
        expect_lb_port = 70
        expect_lb_algorithm = "RANDOM"
        expect_lb_httpsRedirect = "redirect"
        expect_lb_halfClosed = "halfClosed"
        expect_lb_connectionLogging = {"enabled": True}
        expect_lb_contentCaching = True
        expect_lb_timeout = 35

        input_lb_info = {"name": expect_lb_name,
                         "protocol": expect_lb_protocol,
                         "port": expect_lb_port,
                         "timeout": expect_lb_timeout,
                         "httpsRedirect": expect_lb_httpsRedirect,
                         "halfClosed": expect_lb_halfClosed,
                         "connectionLogging": expect_lb_connectionLogging,
                         "contentCaching": expect_lb_contentCaching}

        input_lb_id = "13579"
        input_lb_time = "current_time"
        input_lb_status = "ACTIVE"

        actual = load_balancer_example(input_lb_info, input_lb_id, input_lb_status,
                                       input_lb_time)

        lb_example = {"name": expect_lb_name,
                      "id": input_lb_id,
                      "protocol": expect_lb_protocol,
                      "port": expect_lb_port,
                      "algorithm": expect_lb_algorithm,
                      "status": input_lb_status,
                      "cluster": {"name": "test-cluster"},
                      "timeout": expect_lb_timeout,
                      "created": {"time": input_lb_time},
                      "virtualIps": [{"address": "127.0.0.1",
                                     "id": 1111, "type": "PUBLIC", "ipVersion": "IPV4"},
                                     {"address": "0000:0000:0000:0000:1111:111b:0000:0000",
                                      "id": 1111,
                                      "type": "PUBLIC",
                                      "ipVersion": "IPV6"}],
                      "sourceAddresses": {"ipv6Public": "0000:0001:0002::00/00",
                                          "ipv4Servicenet": "127.0.0.1",
                                          "ipv4Public": "127.0.0.1"},
                      "httpsRedirect": expect_lb_httpsRedirect,
                      "updated": {"time": input_lb_time},
                      "halfClosed": expect_lb_halfClosed,
                      "connectionLogging": expect_lb_connectionLogging,
                      "contentCaching": {"enabled": False}}

        self.assertEqual(actual, lb_example)


@attr.s
class _CLBChangeResponseAndID(object):
    """
    A simple data structure intended to conveniently communicate the results
    of issuing a CLB control plane request.
    """
    resp = attr.ib()
    "The response returned from the .../attributes PATCH request."

    lb_id = attr.ib()
    "The CLB ID used for the purposes of performing the test."


class LoadbalancerAPITests(SynchronousTestCase):
    """
    Tests for the Loadbalancer plugin API
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`LoadBalancerApi` as the only plugin
        """
        lb = LoadBalancerApi()
        self.helper = APIMockHelper(self, [lb, LoadBalancerControlApi(lb_api=lb)])
        self.root = self.helper.root
        self.uri = self.helper.uri

    def _create_loadbalancer(self, name=None, api_helper=None, nodes=None):
        """
        Helper method to create a load balancer and return the lb_id.

        :param str name: The name fo the load balancer, defaults to 'test_lb'
        :param api_helper: An instance of :class:`APIMockHelper` - defaults to
            the one created by setup, but if different regions need to be
            created, for instance, your test may make a different helper, and
            so that helper can be passed here.
        :param list nodes: A list of nodes to create the load balancer with -
            defaults to creating a load balancer with no nodes.

        :return: Load balancer ID
        :rtype: int
        """
        api_helper = api_helper or self.helper
        lb_body = {
            "loadBalancer": {
                "name": name or "test_lb",
                "protocol": "HTTP",
                "virtualIps": [{"type": "PUBLIC"}]
            }
        }
        if nodes is not None:
            lb_body['loadBalancer']['nodes'] = nodes

        resp, body = self.successResultOf(json_request(
            self, api_helper.root, "POST", api_helper.uri + '/loadbalancers',
            lb_body
        ))
        return body['loadBalancer']['id']

    def _patch_attributes_request(
        self, lb_id_offset=0, status_key=None, status_val=None
    ):
        """
        Creates a CLB for the tenant, then attempts to patch its status using
        the CLB control plane endpoint.

        :param int lb_id_offset: Defaults to 0.  If provided, the CLB that is
            created for the tenant will be referenced in the patch request
            offset by this much.
        :param str status_key: Defaults to '"status"'.  If provided, the patch
            will be made against this member of the CLB's state.  Note that
            surrounding quotes are required for th key, thus giving the caller
            the ability to deliberately distort the JSON.
        :param str status_val: Defaults to 'PENDING_DELETE'.  If provided, the
            provided setting will be used for the status key provided.
        :return: An instance of _CLBChangeResponseAndID.  The `resp` attribute
            will refer to Mimic's response object; `code` will be set to the
            HTTP result code from the request.
        """
        ctl_uri = self.helper.auth.get_service_endpoint(
            "cloudLoadBalancerControl", "ORD"
        )
        lb_id = self._create_loadbalancer('test_lb') + lb_id_offset
        status_key = status_key or '"status"'
        status_val = status_val or 'PENDING_DELETE'
        payload = '{{{0}: "{1}"}}'.format(status_key, status_val)
        set_attributes_req = request(
            self, self.root, "PATCH", "{0}/loadbalancer/{1}/attributes".format(
                ctl_uri, lb_id
            ),
            payload
        )
        return _CLBChangeResponseAndID(
            resp=self.successResultOf(set_attributes_req), lb_id=lb_id
        )

    def test_lb_status_changes_as_requested(self):
        """
        Clients can ``PATCH`` to the ``.../loadbalancer/<lb-id>/attributes``
        :obj:`LoadBalancerControlApi` endpoint in order to change the
        ``status`` attribute on the load balancer identified by the given
        load-balancer ID for the same tenant in the :obj:`LoadBalancerApi`.

        This attribute controls the status code returned when the load balancer
        is retrieved by a ``GET`` request.
        """
        r = self._patch_attributes_request()
        self.assertEqual(r.resp.code, 204)

        get_lb = request(self, self.root, "GET", self.uri + '/loadbalancers/' + str(r.lb_id))
        get_lb_response = self.successResultOf(get_lb)
        get_lb_response_body = self.successResultOf(
            treq.json_content(get_lb_response)
        )["loadBalancer"]
        self.assertEqual(get_lb_response.code, 200)
        self.assertEqual(get_lb_response_body["status"], "PENDING_DELETE")

    def test_lb_status_change_with_illegal_json(self):
        """
        In the event the user sends a malformed request body to the
        .../attributes endpoint, we should get back a 400 Bad Request.
        """
        r = self._patch_attributes_request(status_key="\"status'")
        self.assertEqual(r.resp.code, 400)

    def test_lb_status_change_with_bad_keys(self):
        """
        In the event the user sends a request to alter a key which isn't
        supported, we should get back a 400 Bad Request as well.
        """
        r = self._patch_attributes_request(status_key="\"stats\"")
        self.assertEqual(r.resp.code, 400)

    def test_lb_status_change_to_illegal_status(self):
        """
        If we attempt to set a valid status on a valid CLB for a valid tenant
        to a value which is nonsensical, we should get back a 400.
        """
        r = self._patch_attributes_request(status_val="KJDHSFLKJDSH")
        self.assertEqual(r.resp.code, 400)

    def test_lb_status_change_against_undefined_clb(self):
        """
        In the event the user sends a request to alter a key on a load balancer
        which doesn't belong to the requestor, we should get back a 404 code.
        """
        r = self._patch_attributes_request(lb_id_offset=1000)
        self.assertEqual(r.resp.code, 404)

    def test_multiple_regions_multiple_endpoints(self):
        """
        API object created with multiple regions has multiple entries
        in the service catalog.
        """
        helper = APIMockHelper(self,
                               [LoadBalancerApi(regions=['ORD', 'DFW'])])
        entry = helper.service_catalog_json['access']['serviceCatalog'][0]
        self.assertEqual(2, len(entry['endpoints']))

    def test_add_load_balancer(self):
        """
        If created without nodes, no node information appears in the response
        when making a request to ``POST /v1.0/<tenant_id>/loadbalancers``.
        """
        lb_name = 'mimic_lb'
        resp, body = self.successResultOf(json_request(
            self, self.root, "POST", self.uri + '/loadbalancers',
            {
                "loadBalancer": {
                    "name": lb_name,
                    "protocol": "HTTP",
                    "virtualIps": [{"type": "PUBLIC"}]
                }
            }
        ))
        self.assertEqual(resp.code, 202)
        self.assertEqual(body['loadBalancer']['name'], lb_name)
        self.assertNotIn("nodeCount", body["loadBalancer"])
        self.assertNotIn("nodes", body["loadBalancer"])

    def test_add_load_balancer_request_with_no_body_causes_bad_request(self):
        """
        Test to verify :func:`add_load_balancer` on ``POST /v1.0/<tenant_id>/loadbalancers``
        """
        create_lb = request(self, self.root, "POST", self.uri + '/loadbalancers', "")
        create_lb_response = self.successResultOf(create_lb)
        self.assertEqual(create_lb_response.code, 400)

    def test_add_load_balancer_request_with_invalid_body_causes_bad_request(self):
        """
        Test to verify :func:`add_load_balancer` on ``POST /v1.0/<tenant_id>/loadbalancers``
        """
        create_lb = request(self, self.root, "POST", self.uri + '/loadbalancers', "{ bad request: }")
        create_lb_response = self.successResultOf(create_lb)
        self.assertEqual(create_lb_response.code, 400)

    def test_add_load_balancer_with_nodes(self):
        """
        Making a request to ``POST /v1.0/<tenant_id>/loadbalancers`` with
        nodes adds the nodes to the load balancer.
        """
        lb_name = 'mimic_lb'
        resp, body = self.successResultOf(json_request(
            self, self.root, "POST", self.uri + '/loadbalancers',
            {
                "loadBalancer": {
                    "name": lb_name,
                    "protocol": "HTTP",
                    "virtualIps": [{"type": "PUBLIC"}],
                    "nodes": [{"address": "127.0.0.2",
                               "port": 80,
                               "condition": "ENABLED",
                               "type": "PRIMARY"},
                              {"address": "127.0.0.0",
                               "port": 80,
                               "condition": "ENABLED",
                               "type": "SECONDARY"}]
                }
            }
        ))
        self.assertEqual(resp.code, 202)
        self.assertNotIn("nodeCount", body['loadBalancer'])
        self.assertEqual(len(body['loadBalancer']['nodes']), 2)

    def test_list_loadbalancers(self):
        """
        Test to verify :func:`list_load_balancers` with on ``GET /v1.0/<tenant_id>/loadbalancers``
        Create two load balancers, then list them and verify the ids
        """
        test1_id = self._create_loadbalancer('test1')
        test2_id = self._create_loadbalancer('test2')
        list_lb = request(self, self.root, "GET", self.uri + '/loadbalancers')
        list_lb_response = self.successResultOf(list_lb)
        list_lb_response_body = self.successResultOf(treq.json_content(list_lb_response))
        self.assertEqual(list_lb_response.code, 200)
        self.assertEqual(len(list_lb_response_body['loadBalancers']), 2)
        self.assertTrue(list_lb_response_body['loadBalancers'][0]['id'] in [test1_id, test2_id])
        self.assertTrue(list_lb_response_body['loadBalancers'][1]['id'] in [test1_id, test2_id])
        self.assertTrue(list_lb_response_body['loadBalancers'][0]['id'] !=
                        list_lb_response_body['loadBalancers'][1]['id'])

    def test_list_loadbalancers_have_no_nodes(self):
        """
        When listing load balancers, nodes do not appear even if the load
        balanacer has nodes.  "nodeCount" is present for all the load
        balancers, whether or not there are nodes on the load balancer.
        """
        self._create_loadbalancer('no_nodes')
        self._create_loadbalancer(
            '3nodes', nodes=[{"address": "1.1.1.{0}".format(i),
                              "port": 80, "condition": "ENABLED"}
                             for i in range(1, 4)])
        list_resp, list_body = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/loadbalancers'))
        self.assertEqual(list_resp.code, 200)
        self.assertEqual(len(list_body['loadBalancers']), 2)

        for lb in list_body['loadBalancers']:
            self.assertNotIn("nodes", lb)
            self.assertEqual(
                lb['nodeCount'],
                0 if lb['name'] == 'no_nodes' else 3)

    def test_delete_loadbalancer(self):
        """
        Test to verify :func:`delete_load_balancer` with on
        ``DELETE /v1.0/<tenant_id>/loadbalancers/<lb_id>``
        Create two load balancers, then list them and verify the ids
        """
        # These will fail if the servers weren't created
        test1_id = self._create_loadbalancer('test1')
        test2_id = self._create_loadbalancer('test2')
        delete_lb = request(self, self.root, 'DELETE', self.uri + '/loadbalancers/' + str(test1_id))
        del_lb_response = self.successResultOf(delete_lb)
        # This response code does not match the Rackspace documentation which specifies a 200 response
        # See comment: http://bit.ly/1AVHs3v
        self.assertEqual(del_lb_response.code, 202)
        del_lb_response_body = self.successResultOf(treq.content(del_lb_response))
        self.assertEqual(del_lb_response_body, '')
        # List lb to make sure the correct lb is gone and the other remains
        list_lb = request(self, self.root, "GET", self.uri + '/loadbalancers')
        list_lb_response = self.successResultOf(list_lb)
        list_lb_response_body = self.successResultOf(treq.json_content(list_lb_response))
        self.assertTrue(len(list_lb_response_body['loadBalancers']), 1)
        self.assertTrue(list_lb_response_body['loadBalancers'][0]['id'] == test2_id)

    def test_get_loadbalancer_with_nodes(self):
        """
        If there are nodes on the load balancer, "nodes" (but not "nodeCount")
        appears in the response when making a request to
        ``GET /v1.0/<tenant_id>/loadbalancers/<loadbalancer_id>``.
        """
        lb_id = self._create_loadbalancer(
            nodes=[{"address": "1.2.3.4", "port": 80, "condition": "ENABLED"}])
        resp, body = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/loadbalancers/' + str(lb_id)))
        self.assertEqual(resp.code, 200)
        self.assertEqual(body['loadBalancer']['id'], lb_id)
        self.assertNotIn('nodeCount', body['loadBalancer'])
        self.assertEqual(len(body['loadBalancer']['nodes']), 1)

    def test_get_loadbalancer_no_nodes(self):
        """
        If there are no nodes on the load balancer, then neither "nodeCount"
        nor "nodes" appear in the response when making a request to
        ``GET /v1.0/<tenant_id>/loadbalancers/<loadbalancer_id>``
        """
        lb_id = self._create_loadbalancer()
        resp, body = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/loadbalancers/' + str(lb_id)))
        self.assertEqual(resp.code, 200)
        self.assertEqual(body['loadBalancer']['id'], lb_id)
        self.assertNotIn('nodeCount', body['loadBalancer'])
        self.assertNotIn('nodes', body['loadBalancer'])

    def test_get_non_existant_loadbalancer(self):
        """
        Test to verify :func:`get_load_balancers` for a non existant load balancer id.
        """
        get_lb = request(self, self.root, "GET", self.uri + '/loadbalancers/123')
        get_lb_response = self.successResultOf(get_lb)
        self.assertEqual(get_lb_response.code, 404)

    def test_delete_non_existant_loadbalancer(self):
        """
        Test to verify :func:`delete_load_balancers` for a non existant load balancer.
        """
        delete_lb = request(self, self.root, 'DELETE', self.uri + '/loadbalancers/123')
        delete_lb_response = self.successResultOf(delete_lb)
        self.assertEqual(delete_lb_response.code, 404)

    def test_list_loadbalancers_when_none_exist(self):
        """
        Test to verify :func:`list_load_balancers` when no loadbalancers exist.
        """
        list_lb = request(self, self.root, 'GET', self.uri + '/loadbalancers')
        list_lb_response = self.successResultOf(list_lb)
        self.assertEqual(list_lb_response.code, 200)
        list_lb_response_body = self.successResultOf(treq.json_content(list_lb_response))
        self.assertEqual(list_lb_response_body, {"loadBalancers": []})

    def test_different_tenants_same_region_different_lbs(self):
        """
        Creating a LB for one tenant in a particular region should not
        create it for other tenants in the same region.
        """
        self._create_loadbalancer()

        other_tenant = TenantAuthentication(self, self.root, "other", "other")

        list_lb_response, list_lb_response_body = self.successResultOf(
            request_with_content(
                self, self.root, "GET",
                other_tenant.get_service_endpoint("cloudLoadBalancers")
                + "/loadbalancers"))

        self.assertEqual(list_lb_response.code, 200)

        list_lb_response_body = json.loads(list_lb_response_body)
        self.assertEqual(list_lb_response_body, {"loadBalancers": []})

    def test_same_tenant_different_regions(self):
        """
        Creating an LB for a tenant in one different regions should create it
        in another region for that tenant.
        """
        helper = APIMockHelper(self,
                               [LoadBalancerApi(regions=["ORD", "DFW"])])
        self._create_loadbalancer(api_helper=helper)

        list_lb_response, list_lb_response_body = self.successResultOf(
            request_with_content(
                self, helper.root, "GET",
                helper.get_service_endpoint("cloudLoadBalancers", "DFW")
                + "/loadbalancers"))

        self.assertEqual(list_lb_response.code, 200)

        list_lb_response_body = json.loads(list_lb_response_body)
        self.assertEqual(list_lb_response_body, {"loadBalancers": []})


def _bulk_delete(test_case, root, uri, lb_id, node_ids):
    """Bulk delete multiple nodes."""
    query = '?' + '&'.join('id=' + str(node_id) for node_id in node_ids)
    endpoint = uri + '/loadbalancers/' + str(lb_id) + '/nodes' + query
    d = request(test_case, root, "DELETE", endpoint)
    response = test_case.successResultOf(d)
    body = test_case.successResultOf(treq.content(response))
    if body == '':
        body = EMPTY_RESPONSE
    else:
        body = json.loads(body)
    return response, body


def _update_clb_node(test_case, helper, lb_id, node_id, update_data,
                     request_func=json_request):
    """
    Return the response for updating a CLB node.
    """
    return test_case.successResultOf(request_func(
        test_case, helper.root, "PUT",
        "{0}/loadbalancers/{1}/nodes/{2}".format(
            helper.get_service_endpoint("cloudLoadBalancers"),
            lb_id, node_id),
        update_data
    ))


class LoadbalancerNodeAPITests(SynchronousTestCase):
    """
    Tests for the Loadbalancer plugin API for CRUD for nodes.
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`LoadBalancerApi` as the only plugin.
        And create a load balancer and add nodes to the load balancer.
        """
        self.helper = APIMockHelper(self, [LoadBalancerApi()])
        self.root = self.helper.root
        self.uri = self.helper.uri
        create_lb = request(
            self, self.root, "POST", self.uri + '/loadbalancers',
            json.dumps({
                "loadBalancer": {
                    "name": "test_lb",
                    "protocol": "HTTP",
                    "virtualIps": [{"type": "PUBLIC"}]
                }
            })
        )
        create_lb_response = self.successResultOf(create_lb)
        self.create_lb_response_body = self.successResultOf(treq.json_content(
                                                            create_lb_response))
        self.lb_id = self.create_lb_response_body["loadBalancer"]["id"]
        create_node = self._create_nodes(["127.0.0.1"])
        [(self.create_node_response, self.create_node_response_body)] = create_node
        self.node = self.create_node_response_body["nodes"]

    def _create_nodes(self, addresses):
        """
        Create nodes based on the addresses passed.

        :param list addresses: addresses to create nodes for.

        :return: a list of two-tuples of (response, response_body).
        """
        responses = [
            request(
                self, self.root, "POST", self.uri + '/loadbalancers/' +
                str(self.create_lb_response_body["loadBalancer"]["id"]) + '/nodes',
                json.dumps({"nodes": [{"address": address,
                                       "port": 80,
                                       "condition": "ENABLED",
                                       "type": "PRIMARY",
                                       "weight": 10}]}))
            for address in addresses]
        responses = map(self.successResultOf, responses)
        response_bodies = [self.successResultOf(treq.json_content(response))
                           for response in responses]
        return zip(responses, response_bodies)

    def _get_nodes(self, lb_id):
        """Get all the nodes in a LB."""
        list_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(lb_id) + '/nodes')
        response = self.successResultOf(list_nodes)
        body = self.successResultOf(treq.json_content(response))
        return body['nodes']

    def test_add_node_to_loadbalancer(self):
        """
        Test to verify :func: `add_node` create a node successfully.
        """
        self.assertEqual(self.create_node_response.code, 202)
        self.assertEqual(len(self.create_node_response_body["nodes"]), 1)
        # verify that the node has all the attributes
        node1 = self.create_node_response_body["nodes"][0]
        self.assertEqual(node1["status"], "ONLINE")
        self.assertEqual(node1["port"], 80)
        self.assertEqual(node1["type"], "PRIMARY")
        self.assertTrue(node1["id"])
        self.assertEqual(node1["address"], "127.0.0.1")
        self.assertEqual(node1["condition"], "ENABLED")
        self.assertEqual(node1["weight"], 10)

    def test_add_multiple_nodes(self):
        """
        Test to verify :func: `add_node` creates multiple node successfully.
        """
        create_multiple_nodes = request(
            self, self.root, "POST", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes',
            json.dumps({"nodes": [{"address": "127.0.0.2",
                                   "port": 80,
                                   "condition": "ENABLED",
                                   "type": "PRIMARY"},
                                  {"address": "127.0.0.0",
                                   "port": 80,
                                   "condition": "ENABLED",
                                   "type": "SECONDARY"}]})
        )
        create_node_response = self.successResultOf(create_multiple_nodes)
        create_node_response_body = self.successResultOf(treq.json_content(
                                                         create_node_response))
        self.assertEqual(create_node_response.code, 202)
        self.assertEqual(len(create_node_response_body["nodes"]), 2)

    def test_add_duplicate_node(self):
        """
        Test to verify :func: `add_node` does not allow creation of duplicate nodes.
        """
        create_duplicate_nodes = request(
            self, self.root, "POST", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes',
            json.dumps({"nodes": [{"address": "127.0.0.1",
                                   "port": 80,
                                   "condition": "ENABLED",
                                   "type": "PRIMARY"}]})
        )
        create_node_response = self.successResultOf(create_duplicate_nodes)
        self.assertEqual(create_node_response.code, 413)

    def test_add_single_over_node_limit(self):
        """
        Test to verify :func: `add_node` does not allow creation of a single
        node at a time to exceed the node limit.

        Note: This assumes the node limit is 25. If the limit is made
        configurable, this test will need to be updated.
        """

        for port in range(101, 126):
            request(
                self, self.root, "POST", self.uri + '/loadbalancers/' +
                str(self.lb_id) + '/nodes',
                json.dumps({"nodes": [{"address": "127.0.0.1",
                                       "port": port,
                                       "condition": "ENABLED"}]})
            )
        create_over_node = request(
            self, self.root, "POST", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes',
            json.dumps({"nodes": [{"address": "127.0.0.2",
                                   "port": 130,
                                   "condition": "ENABLED",
                                   "type": "SECONDARY"}]})
        )

        create_node_response = self.successResultOf(create_over_node)
        self.assertEqual(create_node_response.code, 413)

    def test_add_bulk_nodes_over_limit(self):
        """
        Test to verify :func: `add_node` does not allow creation of a single
        node at a time to exceed the node limit.

        Note: This assumes the node limit is 25. If the limit is made
        configurable, this test will need to be updated.
        """

        add_node_list = []
        for a in range(26):
            node_addr = "127.0.0.{0}".format(a)
            add_node_list.append({"address": node_addr,
                                  "port": 88,
                                  "condition": "ENABLED",
                                  "type": "SECONDARY"})

        create_over_node = request(
            self, self.root, "POST", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes',
            json.dumps({"nodes": add_node_list})
        )
        create_node_response = self.successResultOf(create_over_node)
        self.assertEqual(create_node_response.code, 413)

    def test_add_node_request_with_no_body_causes_bad_request(self):
        """
        Test to verify :func: `add_node` does not fail on bad request.
        """
        create_duplicate_nodes = request(
            self, self.root, "POST", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes', "")

        create_node_response = self.successResultOf(create_duplicate_nodes)
        self.assertEqual(create_node_response.code, 400)

    def test_add_node_request_with_invalid_body_causes_bad_request(self):
        """
        Test to verify :func: `add_node` does not fail on bad request.
        """
        create_duplicate_nodes = request(
            self, self.root, "POST", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes', "{ bad request: }")

        create_node_response = self.successResultOf(create_duplicate_nodes)
        self.assertEqual(create_node_response.code, 400)

    def test_add_node_to_non_existant_loadbalancer(self):
        """
        Test to verify :func: `add_node` does not allow creation of nodes
        on non existant load balancers.
        """
        create_duplicate_nodes = request(
            self, self.root, "POST", self.uri + '/loadbalancers/123/nodes',
            json.dumps({"nodes": [{"address": "127.0.0.1",
                                   "port": 80,
                                   "condition": "ENABLED",
                                   "type": "PRIMARY"}]})
        )
        create_node_response = self.successResultOf(create_duplicate_nodes)
        self.assertEqual(create_node_response.code, 404)

    def test_list_nodes_on_loadbalancer(self):
        """
        Test to verify :func: `list_node` lists the nodes on the loadbalancer.
        """
        list_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes')
        list_nodes_response = self.successResultOf(list_nodes)
        list_nodes_response_body = self.successResultOf(treq.json_content(
                                                        list_nodes_response))
        self.assertEqual(list_nodes_response.code, 200)
        self.assertEqual(len(list_nodes_response_body["nodes"]), 1)

    def test_list_nodes_on_non_existant_loadbalancer(self):
        """
        Test to verify :func: `list_node` lists the nodes on the loadbalancer.
        """
        list_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/123/nodes')
        list_nodes_response = self.successResultOf(list_nodes)
        self.assertEqual(list_nodes_response.code, 404)

    def test_get_node_on_loadbalancer(self):
        """
        Test to verify :func: `get_node` gets the nodes on the loadbalancer.
        """
        get_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes/'
            + str(self.node[0]["id"]))
        get_node_response = self.successResultOf(get_nodes)
        get_node_response_body = self.successResultOf(treq.json_content(
                                                      get_node_response))
        self.assertEqual(get_node_response.code, 200)
        self.assertEqual(len(self.node), 1)
        self.assertEqual(get_node_response_body["node"]["id"],
                         self.node[0]["id"])

    def test_get_node_on_non_existant_loadbalancer(self):
        """
        Test to verify :func: `get_node` does get a nodes on a
        non existant loadbalancer.
        """
        get_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/123' +
            '/nodes/' + str(self.node[0]["id"]))
        get_node_response = self.successResultOf(get_nodes)
        self.assertEqual(get_node_response.code, 404)

    def test_get_non_existant_node_on_loadbalancer(self):
        """
        Test to verify :func: `get_node` does not get a non existant node.
        """
        get_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes/123')
        get_node_response = self.successResultOf(get_nodes)
        self.assertEqual(get_node_response.code, 404)

    def test_delete_node_on_loadbalancer(self):
        """
        Test to verify :func: `delete_node` deletes the node on the loadbalancer.
        """
        delete_nodes = request(
            self, self.root, "DELETE", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes/'
            + str(self.node[0]["id"]))
        delete_node_response = self.successResultOf(delete_nodes)
        self.assertEqual(delete_node_response.code, 202)

        # assert that it lists correctly after
        list_nodes_resp, list_nodes_body = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes'))
        self.assertEqual(list_nodes_resp.code, 200)
        self.assertEqual(len(list_nodes_body["nodes"]), 0)

    def test_delete_node_on_non_existant_loadbalancer(self):
        """
        Test to verify :func: `delete_node` does delete a nodes on a
        non existant loadbalancer.
        """
        delete_nodes = request(
            self, self.root, "DELETE", self.uri + '/loadbalancers/123' +
            '/nodes/' + str(self.node[0]["id"]))
        delete_node_response = self.successResultOf(delete_nodes)
        self.assertEqual(delete_node_response.code, 404)

    def test_delete_non_existant_node_on_loadbalancer(self):
        """
        Test to verify :func: `delete_node` does not delete a non existant node.
        """
        delete_nodes = request(
            self, self.root, "DELETE", self.uri + '/loadbalancers/' +
            str(self.lb_id) + '/nodes/123')
        delete_node_response = self.successResultOf(delete_nodes)
        self.assertEqual(delete_node_response.code, 404)

    def test_bulk_delete(self):
        """
        Test to verify :func: `delete_nodes` deletes the nodes on the loadbalancer.
        """
        node_results = self._create_nodes(['127.0.0.2', '127.0.0.3', '127.0.0.4'])
        node_ids = [
            node_result['id']
            for (response, body) in node_results
            for node_result in body['nodes']]
        node_ids_to_delete = node_ids[:-1]
        response, body = _bulk_delete(
            self, self.root, self.uri, self.lb_id, node_ids_to_delete)
        self.assertEqual(response.code, 202)
        self.assertEqual(body, EMPTY_RESPONSE)
        remaining_nodes = self._get_nodes(self.lb_id)
        # The one in setUp and the extra one we created are remaining
        self.assertEqual(
            [node['id'] for node in remaining_nodes],
            [self.node[0]['id'], node_ids[-1]])

    def test_bulk_delete_no_nodes(self):
        """
        When deleting multiple nodes but not giving any node IDs, a special
        error is returned.
        """
        response, body = _bulk_delete(
            self, self.root, self.uri, self.lb_id, [])
        self.assertEqual(response.code, 400)
        self.assertEqual(
            body,
            {'code': 400,
             'message': "Must supply one or more id's to process this request."})

    def test_bulk_delete_no_nodes_invalid_lb(self):
        """
        When trying to delete multiple nodes from a non-existent LB, the error
        for an empty node list takes precedence over the error for a
        non-existing LB.
        """
        lb_id = self.lb_id + 1
        response, body = _bulk_delete(self, self.root, self.uri, lb_id, [])
        self.assertEqual(response.code, 400)
        self.assertEqual(
            body,
            {'code': 400,
             'message': "Must supply one or more id's to process this request."})

    def test_bulk_delete_invalid_lb(self):
        """
        Bulk-deleting nodes from a non-existent LB returns a 404 and an appropriate
        message.
        """
        lb_id = self.lb_id + 1
        node_ids_to_delete = [self.node[0]['id']]
        response, body = _bulk_delete(self, self.root, self.uri, lb_id, node_ids_to_delete)
        self.assertEqual(response.code, 404)
        self.assertEqual(
            body,
            {'code': 404,
             'message': "Load balancer not found"})

    def test_bulk_delete_nonexistent_nodes(self):
        """
        When trying to delete multiple nodes, if any of the nodes don't exist, no
        nodes are deleted and a special error result is returned.
        """
        node_ids_to_delete = [self.node[0]['id'], 1000000, 1000001]
        response, body = _bulk_delete(
            self, self.root, self.uri, self.lb_id, node_ids_to_delete)
        self.assertEqual(response.code, 400)
        self.assertEqual(
            body,
            {
                "validationErrors": {
                    "messages": [
                        "Node ids 1000000,1000001 are not a part of your loadbalancer"
                    ]
                },
                "message": "Validation Failure",
                "code": 400,
                "details": "The object is not valid"
            }
        )
        # and the one valid node that we tried to delete is still there
        remaining = [node['id'] for node in self._get_nodes(self.lb_id)]
        self.assertEquals(remaining, [self.node[0]['id']])

    def test_updating_node_invalid_json(self):
        """
        When updating a node, if invalid JSON is provided (both actually not
        valid JSON and also not conforming to the schema), a 400 invalid
        JSON error will be returned.  This takes precedence over whether or not
        a load balancer or node actually exists, and precedence over
        validation errors.
        """
        real_lb_id = self.lb_id
        real_node_id = self.node[0]['id']
        fake_lb_id = real_lb_id + 1
        fake_node_id = real_node_id + 1

        combos = ((real_lb_id, real_node_id),
                  (real_lb_id, fake_node_id),
                  (fake_lb_id, fake_node_id))

        invalids = (
            {"node": {"weight": 1, "hockey": "stick"}},
            {"node": {"weight": 1, "status": "OFFLINE"}},
            {"node": {"weight": 1},
             "other": "garbage"},
            {"node": []},
            {"node": 1},
            {"nodes": {"weight": 1}},
            [],
            "not JSON",
            {"node": {"weight": "not a number", "address": "1.1.1.1"}},
            {"node": {"condition": "INVALID", "id": 1}},
            {"node": {"type": "INVALID", "weight": 1000}},
            {"node": {"weight": "not a number", "port": 80}}
        )

        expected = invalid_json_schema()

        for lb_id, node_id in combos:
            for invalid in invalids:
                resp, body = _update_clb_node(
                    self, self.helper, lb_id, node_id, invalid)
                self.assertEqual(
                    (body, resp.code), expected,
                    "{0} should have returned invalid JSON error".format(
                        invalid))

                self.assertEqual(
                    self._get_nodes(real_lb_id), self.node)

    def test_updating_node_validation_error(self):
        """
        When updating a node, if the address or port are provided,
        a 400 validation error will be returned because those are immutable.
        If the weight is <1 or >100, a 400 validation will also be returned.

        These takes precedence over whether or not a load balancer or node
        actually exists.  The error message also contains a list of all the
        validation failures.
        """
        real_lb_id = self.lb_id
        real_node_id = self.node[0]['id']
        fake_lb_id = real_lb_id + 1
        fake_node_id = real_node_id + 1

        combos = ((real_lb_id, real_node_id),
                  (real_lb_id, fake_node_id),
                  (fake_lb_id, fake_node_id))

        for lb_id, node_id in combos:
            data = {"node": {"weight": 1000, "address": "1.1.1.1",
                             "port": 80, "type": "PRIMARY", "id": 12345}}
            for popoff in (None, "address", "port", "weight"):
                if popoff:
                    del data["node"][popoff]

                resp, body = _update_clb_node(
                    self, self.helper, lb_id, node_id, data)
                actual = (body, resp.code)
                expected = updating_node_validation_error(
                    address="address" in data["node"],
                    port="port" in data["node"],
                    weight="weight" in data["node"],
                    id=True)  # id is always there
                self.assertEqual(
                    actual, expected,
                    "Input of {0}.\nGot: {1}\nExpected: {2}".format(
                        data,
                        json.dumps(actual, indent=2),
                        json.dumps(expected, indent=2)))

                self.assertEqual(
                    self._get_nodes(real_lb_id), self.node)

    def test_updating_node_checks_for_invalid_loadbalancer_id(self):
        """
        If the input is valid, but the load balancer ID does not exist,
        a 404 error is returned.
        """
        resp, body = _update_clb_node(
            self, self.helper, self.lb_id + 1, 1234, {"node": {"weight": 1}})

        self.assertEqual((body, resp.code), loadbalancer_not_found())
        self.assertEqual(self._get_nodes(self.lb_id), self.node)

    def test_updating_node_checks_for_invalid_node_id(self):
        """
        If the input is valid, but the node ID does not exist, a 404 error is
        returned.
        """
        resp, body = _update_clb_node(
            self, self.helper, self.lb_id, self.node[0]["id"] + 1,
            {"node": {"weight": 1}})

        self.assertEqual((body, resp.code), node_not_found())
        self.assertEqual(self._get_nodes(self.lb_id), self.node)

    def test_updating_node_success(self):
        """
        Updating a node successfully changes its values.  The response from a
        successful change is just the values that changed.  The body is an
        empty string. It also updates the atom feed of the node and returns
        that when GETing ../loadbalancers/lbid/nodes/nodeid.atom
        """
        original = self.node[0]
        expected = original.copy()
        change = {
            "condition": "DISABLED",
            "weight": 100,
            "type": "SECONDARY"
        }
        expected.update(change)
        # sanity check to make sure we're actually changing stuff
        self.assertTrue(all([change[k] != original[k] for k in change.keys()]))
        resp, body = _update_clb_node(
            self, self.helper, self.lb_id, self.node[0]["id"],
            json.dumps({"node": change}), request_func=request_with_content)
        self.assertEqual(resp.code, 202)
        self.assertEqual(body, "")

        self.assertEqual(self._get_nodes(self.lb_id)[0], expected)

        # check if feed is updated
        d = request(
            self, self.root, "GET",
            "{0}/loadbalancers/{1}/nodes/{2}.atom".format(self.uri, self.lb_id,
                                                          self.node[0]["id"]))
        feed_response = self.successResultOf(d)
        self.assertEqual(feed_response.code, 200)
        self.assertEqual(
            self.successResultOf(treq.content(feed_response)),
            ("<feed xmlns=\"http://www.w3.org/2005/Atom\"><entry>"
             "<summary>Node successfully updated with address: '127.0.0.1', "
             "port: '80', weight: '100', condition: 'DISABLED'</summary>"
             "<updated>1970-01-01T00:00:00.000000Z</updated></entry></feed>"))

    def test_get_feed_node_404(self):
        """
        Getting feed of non-existent node returns 404 with "Node not found"
        XML
        """
        d = request(
            self, self.root, "GET",
            "{0}/loadbalancers/{1}/nodes/{2}.atom".format(self.uri, self.lb_id, 0))
        feed_response = self.successResultOf(d)
        self.assertEqual(feed_response.code, 404)
        self.assertEqual(
            self.successResultOf(treq.content(feed_response)),
            ('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
             '<itemNotFound xmlns="http://docs.openstack.org/loadbalancers/api/v1.0" code="404">'
             '<message>Node not found</message></itemNotFound>'))

    def test_get_feed_clb_404(self):
        """
        Getting feed of node of non-existent CLB returns 404 with
        "load balancer not found" XML
        """
        d = request(
            self, self.root, "GET",
            "{0}/loadbalancers/{1}/nodes/{2}.atom".format(self.uri, 0, 0))
        feed_response = self.successResultOf(d)
        self.assertEqual(feed_response.code, 404)
        self.assertEqual(
            self.successResultOf(treq.content(feed_response)),
            ('<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
             '<itemNotFound xmlns="http://docs.openstack.org/loadbalancers/api/v1.0" code="404">'
             '<message>Load balancer not found</message></itemNotFound>'))


class LoadbalancerAPINegativeTests(SynchronousTestCase):
    """
    Tests for the Loadbalancer plugin API for error injection
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`LoadBalancerApi` as the only plugin
        """
        helper = APIMockHelper(self, [LoadBalancerApi()])
        self.root = helper.root
        self.uri = helper.uri
        self.helper = helper

    def _create_loadbalancer_for_given_metadata(self, metadata=None):
        """
        Helper method to create a load balancer with the given metadata
        """
        create_lb = request(
            self, self.root, "POST", self.uri + '/loadbalancers',
            json.dumps({
                "loadBalancer": {
                    "name": "test_lb",
                    "protocol": "HTTP",
                    "virtualIps": [{"type": "PUBLIC"}],
                    "metadata": metadata or []
                }
            })
        )
        create_lb_response = self.successResultOf(create_lb)
        return create_lb_response

    def _add_node_to_lb(self, lb_id):
        """
        Adds a node to the load balancer and returns the response object
        """
        create_node = request(
            self, self.root, "POST", self.uri + '/loadbalancers/'
            + str(lb_id) + '/nodes',
            json.dumps({"nodes": [{"address": "127.0.0.1",
                                   "port": 80,
                                   "condition": "ENABLED",
                                   "type": "PRIMARY"}]})
        )
        create_node_response = self.successResultOf(create_node)
        return create_node_response

    def _get_loadbalancer(self, lb_id):
        """
        Makes the `GET` call for the given loadbalancer id and returns the
        load balancer object.
        """
        get_lb = request(self, self.root, "GET", self.uri + '/loadbalancers/' + str(lb_id))
        get_lb_response = self.successResultOf(get_lb)
        get_lb_response_body = self.successResultOf(treq.json_content(get_lb_response))
        return get_lb_response_body

    def _delete_loadbalancer(self, lb_id):
        """
        Deletes the given load balancer id and returns the response
        """
        delete_lb = request(self, self.root, "DELETE", self.uri + '/loadbalancers/' +
                            str(lb_id))
        return self.successResultOf(delete_lb)

    def _create_loadbalancer(self, metadata):
        """Create a load balancer and return the response body."""
        create_response = self._create_loadbalancer_for_given_metadata(metadata)
        self.assertEqual(create_response.code, 202)
        create_lb_response_body = self.successResultOf(treq.json_content(create_response))
        lb = create_lb_response_body["loadBalancer"]
        self.assertEqual(lb["status"], "ACTIVE")
        return lb

    def test_create_load_balancer_in_building_state(self):
        """
        Test to verify the created load balancer remains in building
        state for the time is seconds specified in the metadata.
        Adding a node to a lb in BUILD status results in 422.
        """
        metadata = [{"key": "lb_building", "value": 1}]
        create_response = self._create_loadbalancer_for_given_metadata(metadata)
        self.assertEqual(create_response.code, 202)
        create_lb_response_body = self.successResultOf(treq.json_content(create_response))
        lb = create_lb_response_body["loadBalancer"]
        self.assertEqual(lb["status"], "BUILD")
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 422)

    def test_load_balancer_goes_into_error_state_when_adding_node(self):
        """
        Test to verify a load balancer goes into error state when adding a node.
        Adding a node to a loadbalancer in ERROR state results in 422.
        And such a load balancer can only be deleted.
        """
        metadata = [{"key": "lb_error_state", "value": "error"}]
        lb = self._create_loadbalancer(metadata)
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 202)
        # get loadbalncer after adding node and verify its in error state
        errored_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(errored_lb["loadBalancer"]["status"], "ERROR")
        # adding another node to a lb in ERROR state, results in 422
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 422)
        # An lb in ERROR state can be deleted
        delete_lb = request(self, self.root, "DELETE", self.uri + '/loadbalancers/' +
                            str(lb["id"]))
        delete_lb_response = self.successResultOf(delete_lb)
        self.assertEqual(delete_lb_response.code, 202)

    def test_load_balancer_goes_into_pending_update_state(self):
        """
        Test to verify a load balancer goes into PENDING-UPDATE state, for
        the given time in seconds when any action other than DELETE is performed
        on the lb.
        Adding a node to a loadbalancer in PENDING-UPDATE state results in 422.
        And such a load balancer can be deleted.
        """
        metadata = [{"key": "lb_pending_update", "value": 30}]
        lb = self._create_loadbalancer(metadata)
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 202)
        # get loadbalncer after adding node and verify its in PENDING-UPDATE state
        errored_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(errored_lb["loadBalancer"]["status"], "PENDING-UPDATE")
        # Trying to add/list/delete node on a lb in PENDING-UPDATE state, results in 422
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 422)
        delete_nodes = request(
            self, self.root, "DELETE", self.uri + '/loadbalancers/' +
            str(lb["id"]) + '/nodes/123')
        self.assertEqual(self.successResultOf(delete_nodes).code, 422)
        # An lb in PENDING-UPDATE state can be deleted
        delete_lb = request(self, self.root, "DELETE", self.uri + '/loadbalancers/' +
                            str(lb["id"]))
        delete_lb_response = self.successResultOf(delete_lb)
        self.assertEqual(delete_lb_response.code, 202)

    def test_load_balancer_reverts_from_pending_update_state(self):
        """
        Test to verify a load balancer goes into PENDING-UPDATE state, for
        the given time in seconds.
        """
        metadata = [{"key": "lb_pending_update", "value": 1}]
        lb = self._create_loadbalancer(metadata)
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 202)
        # get loadbalncer after adding node and verify its in PENDING-UPDATE state
        errored_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(errored_lb["loadBalancer"]["status"], "PENDING-UPDATE")
        self.helper.clock.advance(1.0)
        # get loadbalncer after adding node and verify its in ACTIVE state
        errored_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(errored_lb["loadBalancer"]["status"], "ACTIVE")

    def test_delete_load_balancer_and_pending_delete_state(self):
        """
        Test to verify a load balancer goes into PENDING-DELETE state, for
        the given time in seconds and then goes into a DELETED status.
        Also, verify when a load balancer in PENDING-DELETE or DELETED status
        is deleted, response code 400 is returned.
        """
        metadata = [{"key": "lb_pending_delete", "value": 1}]
        lb = self._create_loadbalancer(metadata)

        # Verify the lb status goes into PENDING-DELETE
        del_lb_response = self._delete_loadbalancer(lb["id"])
        self.assertEqual(del_lb_response.code, 202)
        del_lb_content = self.successResultOf(treq.content(del_lb_response))
        self.assertEqual(del_lb_content, '')
        deleted_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(deleted_lb["loadBalancer"]["status"], "PENDING-DELETE")

        # Trying to delete a lb in PENDING-DELETE status results in 400
        self.assertEqual(self._delete_loadbalancer(lb["id"]).code, 400)
        self.helper.clock.advance(1.0000001)

        # Lb goes into DELETED status after time specified in metadata
        deleted_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(deleted_lb["loadBalancer"]["status"], "DELETED")

        # Trying to delete a lb in DELETED status results in 400
        self.assertEqual(self._delete_loadbalancer(lb["id"]).code, 400)

        # GET node on load balancer in DELETED status results in 410
        get_node = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(lb["id"]) + '/nodes/123')
        get_node_response = self.successResultOf(get_node)
        self.assertEqual(get_node_response.code, 410)

        # GET node feed on load balancer in DELETED status results in 410
        node_feed = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' +
            str(lb["id"]) + '/nodes/123.atom')
        node_feed_response = self.successResultOf(node_feed)
        self.assertEqual(node_feed_response.code, 410)

        # List node on load balancer in DELETED status results in 410
        list_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' + str(lb["id"])
            + '/nodes')
        self.assertEqual(self.successResultOf(list_nodes).code, 410)

        # Progress past "deleting now"
        self.helper.clock.advance(4000)
        list_nodes = request(
            self, self.root, "GET", self.uri + '/loadbalancers/' + str(lb["id"])
            + '/nodes')
        self.assertEqual(self.successResultOf(list_nodes).code, 404)

    def test_bulk_delete_empty_list_takes_precedence_over_immutable(self):
        """
        When bulk deleting no nodes, the error indicating nodes must be specified
        is returned even when the LB is not ACTIVE.
        """
        metadata = [{"key": "lb_pending_update", "value": 30}]
        lb = self._create_loadbalancer(metadata)

        # Add a node, which should put it into PENDING-UPDATE
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 202)
        updated_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(updated_lb["loadBalancer"]["status"], "PENDING-UPDATE")

        # Now, trying to bulk-delete an empty list of nodes will still return
        # the empty-nodes error.
        response, body = _bulk_delete(self, self.root, self.uri, lb['id'], [])
        self.assertEqual(response.code, 400)
        self.assertEqual(
            body,
            {'code': 400,
             'message': "Must supply one or more id's to process this request."})

    def test_bulk_delete_not_active(self):
        """
        When bulk deleting nodes while the LB is not ACTIVE, a special error is
        returned, even when some of the nodes are invalid.
        """
        metadata = [{"key": "lb_pending_update", "value": 30}]
        lb = self._create_loadbalancer(metadata)

        # Add a node, which should put it into PENDING-UPDATE
        create_node_response = self._add_node_to_lb(lb["id"])
        self.assertEqual(create_node_response.code, 202)
        updated_lb = self._get_loadbalancer(lb["id"])
        self.assertEqual(updated_lb["loadBalancer"]["status"], "PENDING-UPDATE")

        # Now, trying to bulk-delete nodes (including invalid ones) will cause
        # it to return the special error
        response, body = _bulk_delete(
            self, self.root, self.uri, lb['id'], [100, 200])
        self.assertEqual(response.code, 422)
        self.assertEqual(
            body,
            {u'message': u'LoadBalancer is not ACTIVE', u'code': 422})

    def test_updating_node_loadbalancer_state(self):
        """
        If the load balancer is not active, when updating a node a 422 error
        is returned.
        """
        metadata = [{"key": "lb_pending_update", "value": 30}]
        lb_id = self._create_loadbalancer(metadata)["id"]

        # Add a node, which should put it into PENDING-UPDATE
        create_node_response = self._add_node_to_lb(lb_id)
        self.assertEqual(create_node_response.code, 202)
        updated_lb = self._get_loadbalancer(lb_id)
        self.assertEqual(updated_lb["loadBalancer"]["status"],
                         "PENDING-UPDATE")
        node = self.successResultOf(treq.json_content(create_node_response))
        node_id = node["nodes"][0]["id"]

        resp, body = _update_clb_node(self, self.helper, lb_id, node_id,
                                      {"node": {"weight": 1}})

        self.assertEqual((body, resp.code),
                         considered_immutable_error("PENDING-UPDATE", lb_id))



# -*- test-case-name: mimic.test.test_glance -*-
"""
Defines a list of images from glance
"""

import json
from uuid import uuid4
from six import text_type
from zope.interface import implementer
from twisted.plugin import IPlugin
from mimic.canned_responses.glance import get_images, get_image_schema
from mimic.rest.mimicapp import MimicApp
from mimic.catalog import Entry
from mimic.catalog import Endpoint
from mimic.imimic import IAPIMock


@implementer(IAPIMock, IPlugin)
class GlanceApi(object):
    """
    Rest endpoints for mocked Glance Api.
    """

    def __init__(self, regions=["ORD", "DFW", "IAD"]):
        """
        Create a GlanceApi.
        """
        self._regions = regions

    def catalog_entries(self, tenant_id):
        """
        List catalog entries for the Glance API.
        """
        return [
            Entry(
                tenant_id, "image", "cloudImages",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()), prefix="v2")
                    for region in self._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        return GlanceMock(self, uri_prefix, session_store, region).app.resource()


class GlanceMock(object):
    """
    Glance Mock
    """

    def __init__(self, api_mock, uri_prefix, session_store, name):
        """
        Create a glance region with a given URI prefix.
        """
        self.uri_prefix = uri_prefix
        self._api_mock = api_mock
        self._session_store = session_store
        self._name = name

    app = MimicApp()

    @app.route('/v2/<string:tenant_id>/images', methods=['GET'])
    def get_images(self, request, tenant_id):
        """
        Returns a list of glance images. Currently there is no provision
        for shared versus unshared images in the response
        """
        request.setResponseCode(200)
        return json.dumps(get_images())


class GlanceAdminApi(object):
    """
    Rest endpoints for mocked Glance Admin API.
    """

    app = MimicApp()

    def __init__(self, core):
        """
        :param MimicCore core: The core to which this Glance Admin API will be
        communicating.
        """
        self.core = core

    @app.route('/v2/images', methods=['POST'])
    def create_image(self, request):
        """
        Creates a new image and returns response code 201.
        """
        return json.dumps(self.core.glance_admin_image_store.create_image(request))

    @app.route('/v2/images', methods=['GET'])
    def get_images_for_admin(self, request):
        """
        Returns a list of glance images.
        """
        return json.dumps(self.core.glance_admin_image_store.list_images())

    @app.route('/v2/images/<string:image_id>', methods=['GET'])
    def get_image_for_admin(self, request, image_id):
        """
        Returns image with given `image_id`.
        """
        return json.dumps(self.core.glance_admin_image_store.get_image(
            request,
            image_id))

    @app.route('/v2/images/<string:image_id>', methods=['DELETE'])
    def delete_image(self, request, image_id):
        """
        Deletes the image and returns response code 204.
        """
        return self.core.glance_admin_image_store.delete_image(
            request,
            image_id)

    @app.route('/v2/schemas/image', methods=['GET'])
    def get_image_schema_for_admin(self, request):
        """
        Returns the glance image schema.
        """
        return json.dumps(get_image_schema())



"""
Customer Contacts storage object
"""
from characteristic import attributes, Attribute


@attributes(["tenant_id", "email_address", "role",
             Attribute("first_name", default_value="Test FirstName"),
             Attribute("last_name", default_value="Test LastName")])
class Contact(object):
    """
    A :obj:`Contact` is a representation for each contact for a tenant.
    """

    static_defaults = {
        "firstName": "Pat",
        "lastName": "Example",
        "customerAccountType": "CLOUD",
        "contactNumber": "MIMIC-111-111-111",
        "rcn": "MIMIC-111-111-1111",
        "emailAddresses": {
                    "emailAddress": [
                        {
                            "primary": True,
                            "address": None
                        }
                    ]
        },
        "phoneNumbers": {
            "phoneNumber": [
                        {
                            "number": "55555555555"
                        }
            ]
        },
        "roles": {
            "role": []
        },
        "link": [
            {
                "rel": "via",
                "href": "http://mimic-customer.mimic-customer.com"
            }
        ]
    }

    def generate_contacts(self):
        """
        Long-form JSON-serializable object representation of this contact, as
        returned by a GET.
        """
        template = self.static_defaults.copy()
        template.update({
            "firstName": self.first_name,
            "lastName": self.last_name,
            "customerAccountNumber": self.tenant_id,
            "roles": {
                "role": [self.role]
            },
            "emailAddresses": {
                "emailAddress": [
                    {
                        "primary": True,
                        "address": self.email_address
                    }
                ]
            }
        })
        return template


@attributes([Attribute("contacts_store", default_factory=dict)])
class ContactsStore(object):
    """
    A collection of contact objects for a tenant.
    """

    def add_to_contacts_store(self, tenant_id, contact_list):
        """
        Create a new Contact object for each contact in `contact_list`
        and append it to :obj: `ContactsStore` of the tenant.

        The :obj: `contact_list` is a list of tuple containing the email address,
        and the role for the email address
        """
        if not contact_list:
            self.contacts_store[tenant_id] = []
            return
        self.contacts_store[tenant_id] = [
            Contact(tenant_id=tenant_id, email_address=each_contact[0],
                    role=each_contact[1]) for each_contact in contact_list]
        return

    def list_contacts_for_tenant(self, tenant_id):
        """
        Returns the list of contacts for a tenent
        """
        if tenant_id not in self.contacts_store:
            default_contact_list = [('example@example.com', 'TECHNICAL'),
                                    ('example2@example.com', 'TECHNICAL')]
            self.add_to_contacts_store(tenant_id, default_contact_list)

        contacts = [each.generate_contacts()
                    for each in self.contacts_store[tenant_id]]
        return {
            "link": [
                {
                    "rel": "next",
                    "href": "http://mimic-customer.mimic-customer.com"
                }
            ],
            "contact": contacts
        }



"""
Model objects for mimic flavors.
"""

from characteristic import attributes


@attributes(['flavor_id', 'tenant_id', 'name', 'ram', 'vcpus', 'rxtx', 'disk'])
class Flavor(object):
    """
    A Flavor object
    """

    static_defaults = {
        "swap": "",
        "OS-FLV-EXT-DATA:ephemeral": 0,
    }

    def links_json(self, absolutize_url):
        """
        Create a JSON-serializable data structure describing the links to this
        flavor.
        """
        return [
            {
                "href": absolutize_url("v2/{0}/flavors/{1}"
                                       .format(self.tenant_id, self.flavor_id)),
                "rel": "self"
            },
            {
                "href": absolutize_url("{0}/flavors/{1}"
                                       .format(self.tenant_id, self.flavor_id)),
                "rel": "bookmark"
            }
        ]

    def brief_json(self, absolutize_url):
        """
        Brief JSON-serializable version of this flavor, for the non-details
        list flavors request.
        """
        return {
            "id": self.flavor_id,
            "links": self.links_json(absolutize_url),
            "name": self.name
        }

    def detailed_json(self, absolutize_url):
        """
        Long-form JSON-serializable object representation of this flavor, as
        returned by either a GET on this individual flavor or a member in the
        list returned by the list-details request.
        """
        template = self.static_defaults.copy()
        template.update({
            "id": self.flavor_id,
            "links": self.links_json(absolutize_url),
            "name": self.name,
            "ram": self.ram,
            "vcpus": self.vcpus,
            "rxtx_factor": self.rxtx,
            "disk": self.disk,
            "OS-FLV-WITH-EXT-SPECS:extra_specs": self.extra_specs_json()
        })
        return template


class RackspaceStandardFlavor(Flavor):
    """
    A Rackspace standard flavor object representation
    """
    flavors = {"512MB Standard Instance": {"id": "2", "ram": 512, "vcpus": 1, "rxtx_factor": 80,
                                           "disk": 20},
               "1GB Standard Instance": {"id": "3", "ram": 1024, "vcpus": 1, "rxtx_factor": 120,
                                         "disk": 40},
               "2GB Standard Instance": {"id": "4", "ram": 2048, "vcpus": 2, "rxtx_factor": 240,
                                         "disk": 80},
               "4GB Standard Instance": {"id": "5", "ram": 4096, "vcpus": 2, "rxtx_factor": 400,
                                         "disk": 160},
               "8GB Standard Instance": {"id": "6", "ram": 8192, "vcpus": 4, "rxtx_factor": 600,
                                         "disk": 320},
               "15GB Standard Instance": {"id": "7", "ram": 15360, "vcpus": 6, "rxtx_factor": 800,
                                          "disk": 620},
               "30GB Standard Instance": {"id": "8", "ram": 30720, "vcpus": 8, "rxtx_factor": 1200,
                                          "disk": 1200}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a standard flavor.
        """
        return {
            "class": "standard1",
            "policy_class": "standard_flavor"
        }


class RackspaceComputeFlavor(Flavor):
    """
    A Rackspace compute flavor object representatione
    """
    flavors = {"3.75 GB Compute v1": {"id": "compute1-4", "ram": 3840, "vcpus": 2, "rxtx_factor": 312.5,
                                      "disk": 0},
               "7.5 GB Compute v1": {"id": "compute1-8", "ram": 7680, "vcpus": 4, "rxtx_factor": 625,
                                     "disk": 0},
               "15 GB Compute v1": {"id": "compute1-15", "ram": 15360, "vcpus": 8, "rxtx_factor": 1250,
                                    "disk": 0},
               "30 GB Compute v1": {"id": "compute1-30", "ram": 30720, "vcpus": 16, "rxtx_factor": 2500,
                                    "disk": 0},
               "60 GB Compute v1": {"id": "compute1-60", "ram": 61440, "vcpus": 32, "rxtx_factor": 5000,
                                    "disk": 0}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a compute flavor.
        """
        return {
            "class": "compute1",
            "policy_class": "compute_flavor"
        }


class RackspaceGeneralFlavor(Flavor):
    """
    A Rackspace general flavor object representation
    """
    flavors = {"1 GB General Purpose v1": {"id": "general1-1", "ram": 1024, "vcpus": 1,
                                           "rxtx_factor": 200, "disk": 20},
               "2 GB General Purpose v1": {"id": "general1-2", "ram": 2048, "vcpus": 2,
                                           "rxtx_factor": 400, "disk": 40},
               "4 GB General Purpose v1": {"id": "general1-4", "ram": 4096, "vcpus": 4,
                                           "rxtx_factor": 800, "disk": 80},
               "8 GB General Purpose v1": {"id": "general1-8", "ram": 8192, "vcpus": 8,
                                           "rxtx_factor": 1600, "disk": 160}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a general purpose flavor.
        """
        return {
            "class": "general1",
            "policy_class": "general_flavor"
        }


class RackspaceIOFlavor(Flavor):
    """
    A Rackspace IO flavor object representation
    """
    flavors = {"15 GB I/O v1": {"id": "io1-15", "ram": 15360, "vcpus": 4, "rxtx_factor": 1250,
                                "disk": 40},
               "30 GB I/O v1": {"id": "io1-30", "ram": 30720, "vcpus": 8, "rxtx_factor": 2500,
                                "disk": 40},
               "60 GB I/O v1": {"id": "io1-60", "ram": 61440, "vcpus": 16, "rxtx_factor": 5000,
                                "disk": 40},
               "90 GB I/O v1": {"id": "io1-90", "ram": 92160, "vcpus": 24, "rxtx_factor": 7500,
                                "disk": 40},
               "120 GB I/O v1": {"id": "io1-120", "ram": 122880, "vcpus": 32, "rxtx_factor": 10000,
                                 "disk": 40}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a memory flavor.
        """
        return {
            "class": "io1",
            "policy_class": "io_flavor"
        }


class RackspaceMemoryFlavor(Flavor):
    """
    A Rackspace memory flavor object representation
    """
    flavors = {"15 GB Memory v1": {"id": "memory1-15", "ram": 15360, "vcpus": 2, "rxtx_factor": 625,
                                   "disk": 0},
               "30 GB Memory v1": {"id": "memory1-30", "ram": 30720, "vcpus": 4, "rxtx_factor": 1250,
                                   "disk": 0},
               "60 GB Memory v1": {"id": "memory1-60", "ram": 61440, "vcpus": 8, "rxtx_factor": 2500,
                                   "disk": 0},
               "120 GB Memory v1": {"id": "memory1-120", "ram": 122880, "vcpus": 16, "rxtx_factor": 5000,
                                    "disk": 0},
               "240 GB Memory v1": {"id": "memory1-240", "ram": 245760, "vcpus": 32,
                                    "rxtx_factor": 10000, "disk": 0}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a flavor.
        """
        return {
            "class": "memory1",
            "policy_class": "memory_flavor"
        }


class RackspaceOnMetalFlavor(Flavor):
    """
    A Rackspace onMetal flavor object representation
    """
    flavors = {"OnMetal Compute v1": {"id": "onmetal-compute1", "ram": 32768, "vcpus": 20,
                                      "rxtx_factor": 10000, "disk": 32},
               "OnMetal IO v1": {"id": "onmetal-io1", "ram": 131072, "vcpus": 40,
                                 "rxtx_factor": 10000, "disk": 32},
               "OnMetal Memory v1": {"id": "onmetal-memory1", "ram": 524288, "vcpus": 24,
                                     "rxtx_factor": 10000, "disk": 32}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for an onMetal flavor.
        """
        return {
            "quota_resources": "instances=onmetal-compute-v1-instances,ram=onmetal-compute-v1-ram",
            "class": "onmetal",
            "policy_class": "onmetal_flavor"
        }


class RackspacePerformance1Flavor(Flavor):
    """
    A Rackspace perfomance flavor object representation
    """
    flavors = {"1 GB Performance": {"id": "performance1-1", "ram": 1024, "vcpus": 1, "rxtx_factor": 200,
                                    "disk": 20},
               "2 GB Performance": {"id": "performance1-2", "ram": 2048, "vcpus": 2, "rxtx_factor": 400,
                                    "disk": 40},
               "4 GB Performance": {"id": "performance1-4", "ram": 4096, "vcpus": 4, "rxtx_factor": 800,
                                    "disk": 40},
               "8 GB Performance": {"id": "performance1-8", "ram": 8192, "vcpus": 8, "rxtx_factor": 1600,
                                    "disk": 40}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a performance1 flavor.
        """
        return {
            "class": "performance1",
            "policy_class": "performance_flavor"
        }


class RackspacePerformance2Flavor(Flavor):
    """
    A Rackspace performance flavor object representation
    """
    flavors = {"15 GB Performance": {"id": "performance2-15", "ram": 15360, "vcpus": 4,
                                     "rxtx_factor": 1250, "disk": 40},
               "30 GB Performance": {"id": "performance2-30", "ram": 30720, "vcpus": 8,
                                     "rxtx_factor": 2500, "disk": 40},
               "60 GB Performance": {"id": "performance2-60", "ram": 61440, "vcpus": 16,
                                     "rxtx_factor": 5000, "disk": 40},
               "90 GB Performance": {"id": "performance2-90", "ram": 92160, "vcpus": 24,
                                     "rxtx_factor": 7500, "disk": 40},
               "120 GB Performance": {"id": "performance2-120", "ram": 122880, "vcpus": 32,
                                      "rxtx_factor": 10000, "disk": 40}}

    def extra_specs_json(self):
        """
        Create a JSON-serializable data structure describing
        ``OS-FLV-WITH-EXT-SPECS:extra_specs`` for a performance 2 flavor.
        """
        return {
            "class": "performance2",
            "policy_class": "performance_flavor"
        }



import json
import re

from functools import partial

from twisted.internet.task import Clock
from twisted.trial.unittest import SynchronousTestCase

# We can just import from twisted.logger once mimic drops support for
# twisted < 15.2.0
try:
    from twisted.logger.globalLogPublisher import addObserver, removeObserver
except ImportError:
    from twisted.python.log import addObserver, removeObserver

try:
    from twisted.logger import formatEvent as get_log_message
except ImportError:
    from twisted.python.log import textFromEventDict as get_log_message

from mimic.canned_responses.mimic_presets import get_presets
from mimic.core import MimicCore
from mimic.resource import MimicRoot, get_site
from mimic.test.dummy import ExampleAPI
from mimic.test import helpers

json_request = helpers.json_request
request = helpers.request
request_with_content = helpers.request_with_content


def one_api(testCase, core):
    """
    Get the only API registered against a given MimicCore.

    :return: 2-tuple of ``region`` and ``service_id``
    """
    service_id, api = core._uuid_to_api.items()[0]
    region = api.catalog_entries(tenant_id=None)[0].endpoints[0].region
    return (region, service_id)


class ServiceResourceTests(SynchronousTestCase):
    """
    Tests for ``/service/*`` endpoints, handled by
    :func:`MimicRoot.get_service_resource`
    """
    def test_child_resource_gets_base_uri_from_request(self):
        """
        Whatever the URI is used to access mimic is the one that is passed
        back to the plugin when
        :func:`mimic.imimic.IAPMock.resource_for_region` is called.
        """
        example = ExampleAPI()

        core = MimicCore(Clock(), [example])
        root = MimicRoot(core).app.resource()

        # get the region and service id registered for the example API
        (region, service_id) = one_api(self, core)

        request(
            self, root, "GET",
            "http://mybase/mimicking/{0}/{1}/more/stuff".format(service_id, region)
        )

        self.assertEqual(
            "http://mybase/mimicking/{0}/{1}/".format(service_id, region),
            example.store['uri_prefix'])

    def test_service_endpoint_returns_404_if_wrong_service_id(self):
        """
        When the URI used to access a real service has the right region but
        wrong service ID, a 404 is returned and the resource for the service
        is accessed.
        """
        example = ExampleAPI()

        core = MimicCore(Clock(), [example])
        root = MimicRoot(core).app.resource()

        # get the region and service id registered for the example API
        (region, service_id) = one_api(self, core)

        response = self.successResultOf(request(
            self, root, "GET",
            "http://mybase/mimicking/not_{0}/{1}".format(service_id, region)
        ))
        self.assertEqual(404, response.code)
        self.assertEqual([], example.store.keys())

    def test_service_endpoint_returns_404_if_wrong_region(self):
        """
        When the URI used to access a real service has the right service ID
        but wrong service ID, a 404 is returned and the resource for the
        service is accessed.
        """
        example = ExampleAPI()

        core = MimicCore(Clock(), [example])
        root = MimicRoot(core).app.resource()

        # get the region and service id registered for the example API
        (region, service_id) = one_api(self, core)

        response = self.successResultOf(request(
            self, root, "GET",
            "http://mybase/mimicking/not_{0}/{1}".format(service_id, region)
        ))
        self.assertEqual(404, response.code)
        self.assertEqual([], example.store.keys())

    def test_service_endpoint_returns_service_resource(self):
        """
        When the URI used to access a real service has the right service ID
        and right region, the service's resource is used to respond to the
        request.
        """
        core = MimicCore(Clock(), [ExampleAPI('response!')])
        root = MimicRoot(core).app.resource()

        # get the region and service id registered for the example API
        (region, service_id) = one_api(self, core)

        (response, content) = self.successResultOf(request_with_content(
            self, root, "GET",
            "http://mybase/mimicking/{0}/{1}".format(service_id, region)
        ))
        self.assertEqual(200, response.code)
        self.assertEqual('response!', content)


class RootAndPresetTests(SynchronousTestCase):
    """
    Tests for ``/`` (handled by :func:`MimicRoot.help`) and
    ``/mimic/v1.0/presets`` (handled by :func:`MimicRoot.get_mimic_presets`)

    These are placeholder tests because these two endpoints don't do much yet.
    """
    def test_root(self):
        """
        The root (``/``, handled by :func:`MimicRoot.help`) has a bit of help
        text pointing to the identity endpoint
        """
        core = MimicCore(Clock(), [])
        root = MimicRoot(core).app.resource()

        (response, content) = self.successResultOf(request_with_content(
            self, root, 'GET', '/'))
        self.assertEqual(200, response.code)
        self.assertEqual(['text/plain'],
                         response.headers.getRawHeaders('content-type'))
        self.assertIn(' POST ', content)
        self.assertIn('/identity/v2.0/tokens', content)

    def test_presets(self):
        """
        ``/mimic/v1.0/presets`` (handled by
        :func:`MimicRoot.get_mimic_presets`), returns the presets as a JSON
        body with the right headers.
        """
        core = MimicCore(Clock(), [])
        root = MimicRoot(core).app.resource()

        (response, json_content) = self.successResultOf(json_request(
            self, root, 'GET', '/mimic/v1.0/presets'))
        self.assertEqual(200, response.code)
        self.assertEqual(['application/json'],
                         response.headers.getRawHeaders('content-type'))
        self.assertEqual(get_presets, json_content)

    def test_tick(self):
        """
        ``/mimic/v1.1/tick`` (handled by :func:`MimicRoot.advance_time`)
        advances the clock associated with the service.
        """
        clock = Clock()

        def do():
            do.done = True

        do.done = False
        clock.callLater(3.5, do)
        core = MimicCore(clock, [])
        root = MimicRoot(core, clock).app.resource()
        self.assertEqual(do.done, False)
        jreq = json_request(
            self, root, "POST", "/mimic/v1.1/tick", body={"amount": 3.6}
        )
        [response, json_content] = self.successResultOf(jreq)
        self.assertEqual(response.code, 200)
        expected = {
            'advanced': 3.6,
            'now': '1970-01-01T00:00:03.600000Z',
        }
        self.assertEqual(json_content, expected)
        self.assertEqual(do.done, True)

    def test_fastly(self):
        """
        The /fastly pointing to the fastly endpoint
        """
        core = MimicCore(Clock(), [])
        root = MimicRoot(core).app.resource()

        (response, json_content) = self.successResultOf(json_request(
            self, root, 'GET', '/fastly'))
        self.assertEqual(200, response.code)
        self.assertEqual(json_content, {'status': 'ok'})

    def test_send_grid(self):
        """
        ``/sendgrid/mail.send.json`` returns response code 200.
        """
        core = MimicCore(Clock(), [])
        root = MimicRoot(core).app.resource()

        response = self.successResultOf(request(
            self, root, "POST", "/sendgrid/mail.send.json"))
        self.assertEqual(200, response.code)


class RequestTests(SynchronousTestCase):
    """
    Tests for :obj:`mimic.resource.MimicRequest` and
    :obj:`mimic.resource.MimicRequest`, and :obj:`mimic.resource.get_site`.
    """
    def make_request_to_site(self):
        """
        Make a request and return the response.
        """
        core = MimicCore(Clock(), [ExampleAPI('response!')])
        root = MimicRoot(core).app.resource()

        # get the region and service id registered for the example API
        (region, service_id) = one_api(self, core)
        url = "/mimicking/{0}/{1}".format(service_id, region)
        response = self.successResultOf(request(
            self, root, "GET", url, headers={"one": ["two"]}
        ))
        return (response, url)

    def test_default_content_type(self):
        """
        The default content type of all Mimic responses is application/json.
        """
        response, _ = self.make_request_to_site()
        self.assertEqual(['application/json'],
                         response.headers.getRawHeaders('content-type'))

    def test_verbose_logging(self):
        """
        If verbose logging is turned on, the full request and response is
        logged.
        """
        self.patch(helpers, 'get_site', partial(get_site, logging=True))
        logged_events = []
        addObserver(logged_events.append)
        self.addCleanup(removeObserver, logged_events.append)

        response, url = self.make_request_to_site()

        self.assertEqual(2, len(logged_events))
        self.assertTrue(all([not event['isError'] for event in logged_events]))

        messages = [get_log_message(event) for event in logged_events]

        request_match = re.compile(
            "^Received request: GET (?P<url>.+)\n"
            "Headers: (?P<headers>\{.+\})\n\s*$"
        ).match(messages[0])
        self.assertNotEqual(None, request_match)
        self.assertEqual(url, request_match.group('url'))
        headers = json.loads(request_match.group('headers'))
        self.assertEqual(['two'], headers.get('One'))

        response_match = re.compile(
            "^Responding with 200 for: GET (?P<url>.+)\n"
            "Headers: (?P<headers>\{.+\})\n"
            "\nresponse\!\n\s*$"
        ).match(messages[1])
        self.assertNotEqual(None, response_match)
        self.assertEqual(url, response_match.group('url'))
        headers = json.loads(response_match.group('headers'))
        self.assertEqual(['application/json'], headers.get('Content-Type'))



"""
Models relating to identity.
"""
from uuid import uuid4

from attr import Factory, attributes, attr, validators

from six import string_types, text_type

from zope.interface import implementer

from mimic.imimic import ICredential


@attributes
class IdentitySession(object):
    """
    An identity session for a tenant.
    """
    @classmethod
    def from_store(cls, session_store, tenant_id):
        """
        Get the identity session for the given tenant.
        """
        return (session_store
                .session_for_tenant_id(tenant_id)
                .data_for_api(cls, cls))

    identity_admin = attr(default=False,
                          validator=validators.instance_of(bool))


@implementer(ICredential)
@attributes
class PasswordCredentials(object):
    """
    An object representing the required username/password credentials plus
    an optional tenant identifier.

    :ivar username: The username of the user to authenticate.
    :ivar password: The password used to authenticate the user.
    :ivar tenant_id: Optional - the tenant ID of the user.
    """
    username = attr(validator=validators.instance_of(string_types))
    password = attr(validator=validators.instance_of(string_types))
    tenant_id = attr(default=None)
    type_key = "passwordCredentials"

    def get_session(self, session_store):
        """
        :see: :class:`ICredential.get_session`
        """
        return session_store.session_for_username_password(
            self.username, self.password, self.tenant_id)

    @classmethod
    def from_json(cls, json_blob):
        """
        Given an authentication JSON blob, which should look like::

            {
                "auth": {
                    "passwordCredentials": {
                        "username": "user",
                        "password": "pass"
                    },
                    "tenantId": "111111"
                }
            }

        ``"tenantId"`` is interchanable with ``"tenantName"``, and is
        optional.

        :return: a class:`PasswordCredentials` object
        """
        tenant_id = json_blob['auth'].get('tenantName')
        if tenant_id is None:
            tenant_id = json_blob['auth'].get('tenantId')

        username = json_blob['auth'][cls.type_key]['username']
        password = json_blob['auth'][cls.type_key]['password']
        return cls(username=username, password=password, tenant_id=tenant_id)


@implementer(ICredential)
@attributes
class APIKeyCredentials(object):
    """
    An object representing the required username/api-key credentials plus an
    optional tenant identifier.

    :ivar username: The username of the user to authenticate.
    :ivar api_key: The API key used to authenticate the user.
    :ivar tenant_id: Optional - the tenant ID of the user.
    """
    username = attr(validator=validators.instance_of(string_types))
    api_key = attr(validator=validators.instance_of(string_types))
    tenant_id = attr(default=None)
    type_key = "RAX-KSKEY:apiKeyCredentials"

    def get_session(self, session_store):
        """
        :see: :class:`ICredential.get_session`
        """
        return session_store.session_for_api_key(
            self.username, self.api_key, self.tenant_id)

    @classmethod
    def from_json(cls, json_blob):
        """
        Given an authentication JSON blob, which should look like::

            {
                "auth": {
                    "RAX-KSKEY:apiKeyCredentials": {
                        "username": "user",
                        "apiKey": "key"
                    },
                    "tenantId": "111111"
                }
            }

        ``"tenantId"`` is interchanable with ``"tenantName"``, and is
        optional.

        :return: a class:`APIKeyCredentials` object
        """
        tenant_id = json_blob['auth'].get('tenantName')
        if tenant_id is None:
            tenant_id = json_blob['auth'].get('tenantId')

        username = json_blob['auth'][cls.type_key]['username']
        api_key = json_blob['auth'][cls.type_key]['apiKey']
        return cls(username=username, api_key=api_key, tenant_id=tenant_id)


@implementer(ICredential)
@attributes
class TokenCredentials(object):
    """
    An object representing the required token/tenant credentials.

    :ivar token: The auth token to be authenticated
    :ivar tenant_id: The tenant ID the token belongs to.
    """
    token = attr(validator=validators.instance_of(string_types))
    tenant_id = attr(validator=validators.instance_of(string_types))
    type_key = "token"

    def get_session(self, session_store):
        """
        :see: :class:`ICredential.get_session`
        """
        return session_store.session_for_token(self.token, self.tenant_id)

    @classmethod
    def from_json(cls, json_blob):
        """
        Given an authentication JSON blob, which should look like::

            {
                "auth": {
                    "token": {
                        "id": "my_token"
                    },
                    "tenantId": "111111"
                }
            }

        ``"tenantId"`` is interchanable with ``"tenantName"``, and is
        optional.

        :return: a class:`TokenCredentials` object
        """
        tenant_id = json_blob['auth'].get('tenantName')
        if tenant_id is None:
            tenant_id = json_blob['auth'].get('tenantId')

        token = json_blob['auth'][cls.type_key]['id']
        return cls(token=token, tenant_id=tenant_id)


@implementer(ICredential)
@attributes
class ImpersonationCredentials(object):
    """
    An object representing the required
    impersonator-token/impersonated-username credentials.

    :ivar impersonator_token: The auth token of the user requesting an
        impersonation token.
    :ivar impersonated_username: The username of the user to be impersonated.
    :ivar impersonated_token: The impersonation token that can be used to
        authenticate as impersonated user.
    :ivar int expires_in: The number of seconds the impersonation token will
        last.
    """
    impersonator_token = attr()
    impersonated_username = attr(
        validator=validators.instance_of(string_types))
    expires_in = attr(validator=validators.instance_of(int))
    impersonated_token = attr(
        default=Factory(lambda: 'impersonated_token_' + text_type(uuid4())),
        validator=validators.instance_of(string_types))
    type_key = 'RAX-AUTH:impersonation'

    def get_session(self, session_store):
        """
        :see: :class:`ICredential.get_session`
        """
        return session_store.session_for_impersonation(
            self.impersonated_username,
            self.expires_in,
            self.impersonator_token,
            self.impersonated_token)

    @classmethod
    def from_json(cls, json_blob, auth_token):
        """
        Given an impersonation JSON blob, which should look like::

            {
                "RAX-AUTH:impersonation": {
                    "expire-in-seconds": 1000,
                    "user": {
                        "username": "my_user"
                    }
                }
            }

        Along with a header "X-Auth-Token" with the impersonator's token,

        :return: a class:`ImpersonationCredentials` object
        """
        expires_in = json_blob[cls.type_key].get('expire-in-seconds', 86400)
        username = json_blob[cls.type_key]['user']['username']
        return cls(impersonator_token=auth_token,
                   impersonated_username=username,
                   expires_in=int(expires_in))



"""
Tests for L{mimic.tap}
"""

import sys
import types

from zope.interface import implementer

from twisted import plugins

from twisted.internet.interfaces import (
    IStreamServerEndpointStringParser, IStreamServerEndpoint
)
from twisted.internet.defer import succeed

from twisted.plugin import IPlugin
from twisted.python.filepath import FilePath

from twisted.trial.unittest import SynchronousTestCase
from twisted.plugins.mimic import mimicService
from twisted.application.service import IServiceMaker

from mimic.core import MimicCore
from mimic.resource import MimicLoggingRequest, MimicRequest
from mimic.tap import Options, makeService


@implementer(IStreamServerEndpoint)
class FakeEndpoint(object):
    """
    A fake endpoint that records the factories it's listening on.
    """
    def __init__(self):
        """
        Create a list of factories.
        """
        self.factories = []

    def listen(self, factory):
        """
        Store the factory.
        """
        return succeed(self.factories.append(factory))


@implementer(IPlugin)
@implementer(IStreamServerEndpointStringParser)
class FakeEndpointParser(object):
    """
    Fake Endpoint Parser.
    """

    prefix = 'fake'

    def __init__(self):
        """
        Create a list of endpoints.
        """
        self.endpoints = []

    def parseStreamServer(self, reactor, *args, **kwargs):
        """
        Construct a :obj:`FakeEndpoint` and remember it.
        """
        self.endpoints.append(FakeEndpoint())
        return self.endpoints[-1]


def addFakePluginObject(testCase, pluginPackage, pluginObject):
    """
    Add a fake plugin for the duration of the given test.
    """
    dropinName = "a_fake_dropin"
    dropinQualifiedName = pluginPackage.__name__ + "." + dropinName
    module = sys.modules[dropinQualifiedName] = types.ModuleType(
        dropinQualifiedName)
    testCase.addCleanup(lambda: sys.modules.pop(dropinQualifiedName))
    setattr(pluginPackage, dropinName, module)
    testCase.addCleanup(lambda: delattr(pluginPackage, dropinName))
    # Should provide relevant plugin interface, IPlugin
    module.a_plugin = pluginObject
    tempDir = testCase.mktemp()
    fp = FilePath(tempDir)
    fp.createDirectory()
    pluginPackage.__path__.append(tempDir)
    fp.child("a_fake_dropin.py").touch()


class TapTests(SynchronousTestCase):
    """
    Tests for L{mimic.tap}
    """
    def get_site(self, options_list):
        """
        Create the service, and get the site that was created.
        """
        o = Options()
        o.parseOptions(options_list)
        thisFakeParser = FakeEndpointParser()
        addFakePluginObject(self, plugins, thisFakeParser)
        service = makeService(o)
        service.startService()
        self.assertEqual(len(thisFakeParser.endpoints), 1)
        endpoints = thisFakeParser.endpoints[0]
        self.assertEqual(len(endpoints.factories), 1)
        factory = endpoints.factories[0]
        return factory

    def test_listenOption(self):
        """
        Listen options.
        """
        o = Options()
        o.parseOptions(["--listen", "4321"])
        self.assertEqual(o["listen"], "4321")

    def test_makeService(self):
        """
        makeService creates a service that, when listened upon, creates a Site.
        """
        site = self.get_site(["--listen", "fake:"])
        self.assertEqual(site.displayTracebacks, False)
        self.assertEqual(site.requestFactory, MimicRequest)

    def test_realtime(self):
        """
        The C{--realtime} option specifies that the global reactor ought to be
        used, so Mimic's time can be advanced by the passage of real time.
        """
        o = Options()
        o.parseOptions(["--realtime"])

        class CheckClock(MimicCore):
            @classmethod
            def fromPlugins(cls, clock):
                result = super(CheckClock, cls).fromPlugins(clock)
                CheckClock.clock = clock
                return result
        from mimic import tap
        self.patch(tap, "MimicCore", CheckClock)
        makeService(o)
        # Grab the global reactor just for comparison.
        from twisted.internet import reactor as real_reactor
        self.assertIdentical(CheckClock.clock, real_reactor)

    def test_verbose_logging(self):
        """
        The C{--verbose} option causes the default request factory on the site
        to be :class:`MimicLoggingRequest`.
        """
        site = self.get_site(["--listen", "fake:", "--verbose"])
        self.assertEqual(site.displayTracebacks, False)
        self.assertEqual(site.requestFactory, MimicLoggingRequest)

    def test_plugin(self):
        """
        :obj:`twisted.plugins.mimic.mimicService` is a twistd plugin
        referencing `mimic.tap.makeService`.
        """
        self.assertTrue(IPlugin.providedBy(mimicService))
        self.assertTrue(IServiceMaker.providedBy(mimicService))
        self.assertIs(mimicService.makeService, makeService)



"""
Model objects for the Nova mimic.
"""
from characteristic import attributes, Attribute
from uuid import uuid4
from json import loads, dumps

from mimic.util.helper import random_hex_generator


@attributes(["node_id",
             Attribute("chassis_uuid", default_value=None),
             Attribute("driver", default_value=None),
             Attribute("driver_info", default_value=None),
             Attribute("properties", default_value=None),
             Attribute("flavor_id", default_value="onmetal-io1"),
             Attribute("power_state", default_value="power on"),
             Attribute("provision_state", default_value="available"),
             Attribute("instance_uuid", default_value=None),
             Attribute("maintenance", default_value=False),
             Attribute("cache_image_id", default_value=None),
             Attribute("memory_mb", default_value=None),
             Attribute("name", default_value=None)
             ])
class Node(object):
    """
    A :obj:`Node` is a representation of all the state associated with a ironic
    node.  It can produce JSON-serializable objects for various pieces of
    state that are required for API responses.
    """

    static_defaults = {
        "target_power_state": None,
        "target_provision_state": None,
        "updated_at": "2015-08-09T04:30:05+00:00",
        "last_error": None,
        "console_enabled": False,
        "maintenance_reason": None,
        "provision_updated_at": "2015-08-07T06:57:24+00:00",
        "reservation": None,
        "created_at": "2014-09-26T18:56:03+00:00",
        "instance_info": None,
        "inspection_finished_at": None,
        "inspection_started_at": None,
        "clean_step": {},
        "driver_internal_info": {
            "clean_steps": None,
            "hardware_manager_version":
            {
                "generic_hardware_manager": "1",
                "onmetal_hardware_manager": "1"
            },
                "is_whole_disk_image": True,
                "agent_erase_devices_iterations": 1,
                "agent_url": "http://127.0.0.1:8900",
                "cleaning_reboot": True,
                "agent_last_heartbeat": 1440117499
        }
    }

    static_instance_info = {
        "root_gb": "32",
        "image_source": str(uuid4()),
        "ephemeral_gb": "3200",
        "configdrive": str(random_hex_generator(100)),
        "image_url": "http://127.0.0.1/mimic-image-url",
        "image_container_format": "bare_mimic",
        "image_disk_format": "mimic",
        "image_checksum": str(random_hex_generator(6)),
        "swap_mb": "0"
    }

    def links_json(self):
        """
        Create a JSON-serializable data structure describing the links to this
        node.
        """
        return [
            {
                "href": "http://link-to-ironic/v1/nodes/".format(self.node_id),
                "rel": "self"
            },
            {
                "href": "http://link-to-ironic/nodes/".format(self.node_id),
                "rel": "bookmark"
            }
        ]

    def port_links_json(self):
        """
        Create a JSON-serializable data structure describing the port links to this
        node.
        """
        return [
            {
                "href": "http://link-to-ironic/v1/nodes/{0}/ports".format(self.node_id),
                "rel": "self"
            },
            {
                "href": "http://link-to-ironic/nodes/{0}/ports".format(self.node_id),
                "rel": "bookmark"
            }
        ]

    def brief_json(self):
        """
        Brief JSON-serializable version of this server, for the non-details
        list nodes request.
        """
        return {
            "instance_uuid": self.instance_uuid,
            "uuid": self.node_id,
            "links": self.links_json(),
            "maintenance": self.maintenance,
            "provision_state": self.provision_state,
            "power_state": self.power_state,
            "name": self.name
        }

    def detail_json(self):
        """
        Long-form JSON-serializable object representation of this node, as
        returned by either a GET on this individual node or a member in the
        list returned by the list-details request.
        """
        template = self.static_defaults.copy()
        template.update({
            "instance_uuid": self.instance_uuid,
            "chassis_uuid": self.chassis_uuid,
            "name": self.name,
            "uuid": self.node_id,
            "links": self.links_json(),
            "maintenance": self.maintenance,
            "provision_state": self.provision_state,
            "power_state": self.power_state,
            "ports": self.port_links_json(),
            "extra": {
                "flavor": self.flavor_id,
                "hardware/inventory/disks/0/size": "31016853504",
                "hardware/interfaces/0/switch_port_id": "Mimic0/01",
                "rackid": "Mimic_rackid",
                "hardware/interfaces/1/switch_chassis_id": "mimic-chassis1",
                "core_id": "00000",
                "hardware/inventory/disks/2/size": "1861817990000",
                "uutsn": "Mimic-uutsn",
                "hardware/inventory/disks/1/size": "1861817990000",
                "hardware/inventory/cpu/count": "40",
                "hardware/inventory/memory/total": "135234740224",
                "hardware/inventory/disks/2/rotational": "False",
                "hardware/inventory/disks/0/rotational": "False",
                "hardware/inventory/disks/1/rotational": "False",
                "racklocation": "Mimic00",
                "hardware/interfaces/1/switch_port_id": "Mimic1/10",
                "hardware/interfaces/0/switch_chassis_id": "mimic-chassis1"
            },
            "properties": self.properties or {
                "memory_mb": self.memory_mb,
                "cpu_arch": "amd64",
                "local_gb": 32,
                "cpus": 40
            },
            "driver": self.driver or "fake",
            "driver_info": self.driver_info or {
                "ipmi_username": "USERID",
                "ipmi_address": "127.0.0.0",
                "ipmi_password": "******",
                "cache_image_id": self.cache_image_id,
                "cache_status": 'cached' if self.cache_image_id else None
            }
        })
        if self.instance_uuid:
            template["instance_info"] = self.static_instance_info
            template["provision_state"] = "active"
        return template


@attributes([Attribute("ironic_node_store", default_factory=list)])
class IronicNodeStore(object):
    """
    A collection of ironic :obj:`Node` objects.
    """

    memory_to_flavor_map = {131072: "onmetal-io1",
                            32768: "onmetal-compute1",
                            524288: "onmetal-memory1"
                            }

    def node_not_found(self, node_id):
        """
        Error returned by the API calls when the node the action is
        requested upon, does not exist
        """
        return dumps({
            "error_message": {
                "debuginfo": None,
                "faultcode": "Client",
                "faultstring": "Invalid input for field/attribute node_ident. " +
                "Value: 'node={0}'. unable to convert to uuid_or_name".format(node_id)}})

    def node_by_id(self, node_id):
        """
        Retrieve a :obj:`Node` object by its ID.
        """
        for node in self.ironic_node_store:
            if node.node_id == node_id:
                return node

    def add_to_ironic_node_store(self, **attributes):
        """
        Create a new Node object and add it to the
        :obj: `ironic_node_store`
        """
        if not attributes.get('chassis_uuid'):
            attributes['chassis_uuid'] = str(uuid4())
        if (attributes.get("flavor_id", None) is None and
                attributes.get("memory_mb")):
            attributes['flavor_id'] = self.memory_to_flavor_map.get(
                attributes['memory_mb'], "onmetal-mimic")
        node = Node(**attributes)
        self.ironic_node_store.append(node)
        return node

    def create_node(self, http_create_request):
        """
        Create a node
        http://bit.ly/1N0O9KM
        """
        content = loads(http_create_request.content.read())
        try:
            memory_mb = None
            if content.get('properties'):
                memory_mb = content['properties'].get('memory_mb')
            node = self.add_to_ironic_node_store(
                node_id=str(uuid4()),
                memory_mb=memory_mb,
                chassis_uuid=content.get('chassis_uuid'),
                driver=content.get('driver'),
                properties=content.get('properties'),
                driver_info=content.get('driver_info'),
                name=content.get('name'))
            http_create_request.setResponseCode(201)
            return dumps(node.detail_json())
        except:
            http_create_request.setResponseCode(400)
            return b''

    def delete_node(self, http_delete_request, node_id):
        """
        Delete the `node_id` from the :obj:`ironic_node_store` if
        the node exists and set response code to be 204.
        If node does not exist, return response code 404.
        """
        node = self.node_by_id(node_id)
        if node:
            self.ironic_node_store.remove(node)
            http_delete_request.setResponseCode(204)
            return b''
        http_delete_request.setResponseCode(404)
        return b''

    def list_nodes(self, include_details):
        """
        List Ironic nodes.
        If no nodes were added to the :obj:`ironic_node_store`, then adds nodes
        of different onmetal flavors.

        Supports both:
        http://docs.openstack.org/developer/ironic/webapi/v1.html#get--v1-nodes and
        http://docs.openstack.org/developer/ironic/webapi/v1.html#get--v1-nodes-detail
        """
        if not self.ironic_node_store:
            for _ in range(30):
                self.add_to_ironic_node_store(node_id=str(uuid4()),
                                              memory_mb=131072)
            for _ in range(30):
                self.add_to_ironic_node_store(node_id=str(uuid4()),
                                              memory_mb=32768)
            for _ in range(30):
                self.add_to_ironic_node_store(node_id=str(uuid4()),
                                              memory_mb=524288)
            for _ in range(2):
                self.add_to_ironic_node_store(node_id=str(uuid4()),
                                              memory_mb=32768,
                                              instance_uuid=str(uuid4()))
        result = {
            "nodes": [
                node.brief_json() if not include_details
                else node.detail_json()
                for node in self.ironic_node_store
            ]
        }
        return dumps(result)

    def get_node_details(self, http_request, node_id):
        """
        Returns the :obj: `Node` for the corresponding `node_id`
        from the :obj:`ironic_node_store`.
        Returns 404 if `node_id` one does not exist in
        ``self.ironic_node_store``
        Docs:http://bit.ly/1NMIlGx
        """
        if not self.node_by_id(node_id):
            http_request.setResponseCode(404)
            return b''
        return dumps(self.node_by_id(node_id).detail_json())

    def set_node_provision_state(self, http_put_request, node_id):
        """
        Sets the provision state on the node and returns 202.
        If the `node_id` does not exist returns 404.
        Note: When the provision_state is set to `provide` the
        node is made 'available'.
        Docs: http://bit.ly/1ElELdU
        """
        content = loads(http_put_request.content.read())
        node = self.node_by_id(node_id)
        if node:
            http_put_request.setResponseCode(202)
            node.provision_state = content.get('target', 'available')
            if node.provision_state == 'provide':
                node.provision_state = 'available'
            if node.provision_state != 'active':
                node.instance_uuid = None
                node.cache_image_id = None
            if node.provision_state == 'active':
                node.instance_uuid = str(uuid4())
            return dumps(b'')
        http_put_request.setResponseCode(404)
        return self.node_not_found(node_id)

    def cache_image_using_vendor_passthru(self, http_request, node_id, method):
        """
        Cache the image on the node.
        """
        content = loads(http_request.content.read())
        node = self.node_by_id(node_id)
        if not node:
            http_request.setResponseCode(404)
            return self.node_not_found(node_id)
        if method != 'cache_image':
            http_request.setResponseCode(400)
            return b''
        if content.get('image_info') and content['image_info'].get('id'):
            node.cache_image_id = content['image_info']['id']
            http_request.setResponseCode(202)
            return dumps(b'')
        http_request.setResponseCode(400)
        return b''



"""
MaaS API data model
"""

import random
import string
from uuid import uuid4

from characteristic import attributes, Attribute
from six import text_type

from mimic.util.helper import random_hex_generator, random_string

METRIC_TYPE_INTEGER = 'i'
METRIC_TYPE_NUMBER = 'n'
METRIC_TYPE_STRING = 's'


@attributes([Attribute('agent_id', default_value=None),
             Attribute('created_at', instance_of=int),
             Attribute('id',
                       default_factory=(lambda: u'en' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('ip_addresses', default_factory=dict, instance_of=dict),
             Attribute('label', default_value=u'', instance_of=text_type),
             Attribute('managed', default_value=False, instance_of=bool),
             Attribute('metadata', default_factory=dict, instance_of=dict),
             Attribute('updated_at', instance_of=int),
             Attribute('uri', default_value=None)])
class Entity(object):
    """
    Models a MaaS Entity.
    """
    USER_SPECIFIABLE_KEYS = ['agent_id',
                             'ip_addresses',
                             'label',
                             'managed',
                             'metadata',
                             'uri']

    def to_json(self):
        """
        Serializes the Entity to a JSON-encodable dict.
        """
        return {'label': self.label,
                'id': self.id,
                'agent_id': self.agent_id,
                'created_at': self.created_at,
                'updated_at': self.updated_at,
                'managed': self.managed,
                'metadata': self.metadata,
                'ip_addresses': self.ip_addresses,
                'uri': self.uri}

    def update(self, **kwargs):
        """
        Updates this Entity.
        """
        for key in Entity.USER_SPECIFIABLE_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key])
        self.updated_at = int(1000 * kwargs['clock'].seconds())


@attributes([Attribute('created_at', instance_of=int),
             Attribute('details', default_factory=dict, instance_of=dict),
             Attribute('disabled', default_value=False, instance_of=bool),
             Attribute('entity_id', instance_of=text_type),
             Attribute('id',
                       default_factory=(lambda: u'ch' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('label', default_value=u'', instance_of=text_type),
             Attribute('metadata', default_factory=dict, instance_of=dict),
             Attribute('monitoring_zones_poll', default_factory=list, instance_of=list),
             Attribute('period', default_value=60, instance_of=int),
             Attribute('target_alias', default_value=None),
             Attribute('target_hostname', default_value=None),
             Attribute('target_resolver', default_value=None),
             Attribute('timeout', default_value=10, instance_of=int),
             Attribute('type', instance_of=text_type),
             Attribute('updated_at', instance_of=int)])
class Check(object):
    """
    Models a MaaS Check.
    """
    USER_SPECIFIABLE_KEYS = ['details',
                             'disabled',
                             'label',
                             'metadata',
                             'monitoring_zones_poll',
                             'period',
                             'target_alias',
                             'target_hostname',
                             'target_resolver',
                             'timeout',
                             'type']

    def to_json(self):
        """
        Serializes the Check to a JSON-encodable dict.
        """
        return {'label': self.label,
                'id': self.id,
                'type': self.type,
                'monitoring_zones_poll': self.monitoring_zones_poll,
                'created_at': self.created_at,
                'updated_at': self.updated_at,
                'timeout': self.timeout,
                'period': self.period,
                'disabled': self.disabled,
                'metadata': self.metadata,
                'target_alias': self.target_alias,
                'target_resolver': self.target_resolver,
                'target_hostname': self.target_hostname,
                'details': self.details}

    def update(self, **kwargs):
        """
        Updates this Check.
        """
        for key in Check.USER_SPECIFIABLE_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key])
        self.updated_at = int(1000 * kwargs['clock'].seconds())


@attributes([Attribute('check_id', instance_of=text_type),
             Attribute('created_at', instance_of=int),
             Attribute('criteria', default_value=u'', instance_of=text_type),
             Attribute('disabled', default_value=False, instance_of=bool),
             Attribute('entity_id', instance_of=text_type),
             Attribute('id',
                       default_factory=(lambda: u'al' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('label', default_value=u'', instance_of=text_type),
             Attribute('metadata', default_factory=dict, instance_of=dict),
             Attribute('notification_plan_id', instance_of=text_type),
             Attribute('updated_at', instance_of=int)])
class Alarm(object):
    """
    Models a MaaS Alarm.
    """
    USER_SPECIFIABLE_KEYS = ['check_id',
                             'criteria',
                             'disabled',
                             'label',
                             'metadata',
                             'notification_plan_id']

    def to_json(self):
        """
        Serializes the Alarm to a JSON-encodable dict.
        """
        return {'id': self.id,
                'label': self.label,
                'criteria': self.criteria,
                'check_id': self.check_id,
                'entity_id': self.entity_id,
                'notification_plan_id': self.notification_plan_id,
                'created_at': self.created_at,
                'updated_at': self.updated_at,
                'disabled': self.disabled,
                'metadata': self.metadata}

    def update(self, **kwargs):
        """
        Updates this Alarm.
        """
        for key in Alarm.USER_SPECIFIABLE_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key])
        self.updated_at = int(1000 * kwargs['clock'].seconds())


@attributes([Attribute('created_at', instance_of=int),
             Attribute('details', default_factory=dict, instance_of=dict),
             Attribute('id',
                       default_factory=(lambda: u'nt' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('label', default_value=u'', instance_of=text_type),
             Attribute('metadata', default_factory=dict, instance_of=dict),
             Attribute('type', default_value='email', instance_of=text_type),
             Attribute('updated_at', instance_of=int)])
class Notification(object):
    """
    Models a MaaS Notification.
    """
    USER_SPECIFIABLE_KEYS = ['details', 'label', 'metadata', 'type']

    def to_json(self):
        """
        Serializes the Notification to a JSON-encodable dict.
        """
        return {'id': self.id,
                'label': self.label,
                'type': self.type,
                'details': self.details,
                'created_at': self.created_at,
                'updated_at': self.updated_at,
                'metadata': self.metadata}

    def update(self, **kwargs):
        """
        Updates this Notification.
        """
        for key in Notification.USER_SPECIFIABLE_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key])
        self.updated_at = int(1000 * kwargs['clock'].seconds())


@attributes([Attribute('created_at', instance_of=int),
             Attribute('critical_state', default_factory=list, instance_of=list),
             Attribute('id',
                       default_factory=(lambda: u'np' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('label', default_value=u'', instance_of=text_type),
             Attribute('metadata', default_factory=dict, instance_of=dict),
             Attribute('ok_state', default_factory=list, instance_of=list),
             Attribute('updated_at', instance_of=int),
             Attribute('warning_state', default_factory=list, instance_of=list)])
class NotificationPlan(object):
    """
    Models a MaaS notification plan.
    """
    USER_SPECIFIABLE_KEYS = ['critical_state',
                             'label',
                             'metadata',
                             'ok_state',
                             'warning_state']

    def to_json(self):
        """
        Serializes the Notification Plan to a JSON-encodable dict.
        """
        return {'id': self.id,
                'label': self.label,
                'critical_state': self.critical_state,
                'warning_state': self.warning_state,
                'ok_state': self.ok_state,
                'created_at': self.created_at,
                'updated_at': self.updated_at,
                'metadata': self.metadata}

    def update(self, **kwargs):
        """
        Updates this Notification Plan.
        """
        for key in NotificationPlan.USER_SPECIFIABLE_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key])
        self.updated_at = int(1000 * kwargs['clock'].seconds())


@attributes([Attribute('alarms', default_factory=list, instance_of=list),
             Attribute('checks', default_factory=list, instance_of=list),
             Attribute('created_at', instance_of=int),
             Attribute('end_time', default_value=0, instance_of=int),
             Attribute('entities', default_factory=list, instance_of=list),
             Attribute('id',
                       default_factory=(lambda: u'sp' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('label', default_value=u'', instance_of=text_type),
             Attribute('notification_plans', default_factory=list, instance_of=list),
             Attribute('start_time', default_value=0, instance_of=int),
             Attribute('updated_at', instance_of=int)])
class Suppression(object):
    """
    Models a MaaS suppression.
    """
    USER_SPECIFIABLE_KEYS = ['alarms',
                             'checks',
                             'end_time',
                             'entities',
                             'label',
                             'notification_plans',
                             'start_time']

    def to_json(self):
        """
        Serializes the Suppression to a JSON-encodable dict.
        """
        return {'id': self.id,
                'label': self.label,
                'start_time': self.start_time,
                'end_time': self.end_time,
                'notification_plans': self.notification_plans,
                'entities': self.entities,
                'checks': self.checks,
                'alarms': self.alarms}

    def update(self, **kwargs):
        """
        Updates this Suppression.
        """
        for key in Suppression.USER_SPECIFIABLE_KEYS:
            if key in kwargs:
                setattr(self, key, kwargs[key])
        self.updated_at = int(1000 * kwargs['clock'].seconds())


@attributes([Attribute('alarm_changelog_id',
                       default_factory=(lambda: unicode(uuid4())),
                       instance_of=text_type),
             Attribute('alarm_id', instance_of=text_type),
             Attribute('alarm_label', instance_of=text_type),
             Attribute('analyzed_by_monitoring_zone_id', default_value=u'mzord', instance_of=text_type),
             Attribute('check_id', instance_of=text_type),
             Attribute('entity_id', instance_of=text_type),
             Attribute('id',
                       default_factory=(lambda: u'as' + random_hex_generator(4)),
                       instance_of=text_type),
             Attribute('previous_state', instance_of=text_type),
             Attribute('state', instance_of=text_type),
             Attribute('status', instance_of=text_type),
             Attribute('timestamp', instance_of=int)])
class AlarmState(object):
    """
    Models a MaaS alarm state.
    """
    def brief_json(self):
        """
        Serializes this alarm state to a JSON-encodable dict.
        """
        return {'timestamp': self.timestamp,
                'entity_id': self.entity_id,
                'alarm_id': self.alarm_id,
                'check_id': self.check_id,
                'status': self.status,
                'state': self.state,
                'previous_state': self.previous_state,
                'alarm_changelog_id': self.alarm_changelog_id,
                'analyzed_by_monitoring_zone_id': self.analyzed_by_monitoring_zone_id}

    def detail_json(self):
        """
        Serializes this alarm state with additional details.
        """
        details = self.brief_json()
        details.update(alarm_label=self.alarm_label)
        return details


@attributes(["name",
             "type",
             Attribute("unit", default_value="other"),
             Attribute("_overrides", default_factory=dict)])
class Metric(object):
    """
    Models a MaaS metric type.
    """
    def _override_key(self, **kwargs):
        """
        Computes a key used for hashing overrides, as a tuple of strings.
        """
        return (kwargs['entity_id'],
                kwargs['check_id'],
                kwargs.get('monitoring_zone', '__AGENT__'))

    def set_override(self, **kwargs):
        """
        Sets the override metric for a given tenant, entity and check.

        Override metrics are defined as a function which takes a timestamp
        and returns the metric value.
        """
        self._overrides[self._override_key(**kwargs)] = kwargs['override_fn']

    def clear_overrides(self):
        """
        Clears the override metric values.
        """
        self._overrides = {}

    def _get_default_data(self):
        """
        Gets the default data point. This data point may be overridden by
        setting the override value.
        """
        if self.type == METRIC_TYPE_INTEGER:
            return random.randint(0, 100000)
        elif self.type == METRIC_TYPE_NUMBER:
            if self.unit == 'percent':
                return random.uniform(0, 100)
            else:
                return random.uniform(0, 100000)
        elif self.type == METRIC_TYPE_STRING:
            return random_string(random.randint(12, 30), selectable=(string.letters + string.digits))
        raise ValueError('No default data getter for type {0}!'.format(self.type))

    def get_value(self, **kwargs):
        """
        Gets the value of the metric at the specified timestamp.

        Overrides will be applied as necessary.
        """
        override_key = self._override_key(**kwargs)
        timestamp = kwargs['timestamp']
        if override_key in self._overrides:
            return self._overrides[override_key](timestamp)
        return self._get_default_data()

    def get_value_for_test_check(self, **kwargs):
        """
        Gets the metric data object as returned from the test-check API.
        """
        return {'type': self.type,
                'unit': self.unit,
                'data': self.get_value(**kwargs)}


@attributes(["metrics",
             Attribute("_clock"),
             Attribute("test_check_available", default_factory=dict),
             Attribute("test_check_status", default_factory=dict),
             Attribute("test_check_response_code", default_factory=dict)])
class CheckType(object):
    """
    Data model for a MaaS check type (e.g., remote.ping).
    """
    def clear_overrides(self):
        """
        Clears the overrides for test-checks and metrics.
        """
        self.test_check_available = {}
        self.test_check_status = {}
        self.test_check_response_code = {}

        for metric in self.metrics:
            metric.clear_overrides()

    def get_metric_by_name(self, metric_name):
        """
        Gets the metric on this check type.

        This method is useful for setting and clearing overrides on the
        test metrics.
        """
        for metric in self.metrics:
            if metric.name == metric_name:
                return metric
        raise NameError('No metric named "{0}"!'.format(metric_name))

    def get_test_check_response(self, **kwargs):
        """
        Gets the response as would have been returned by the test-check API.
        """
        entity_id = kwargs['entity_id']
        check_id = kwargs.get('check_id', '__test_check')
        monitoring_zones = kwargs.get('monitoring_zones') or ['__AGENT__']

        ench_key = (entity_id, check_id)
        timestamp = int(1000 * self._clock.seconds())

        return (self.test_check_response_code.get(ench_key, 200),
                [{'timestamp': timestamp,
                  'monitoring_zone_id': monitoring_zone,
                  'available': self.test_check_available.get(ench_key, True),
                  'status': self.test_check_status.get(
                      ench_key, 'code=200,rt=0.4s,bytes=99'),
                  'metrics': dict([(m.name,
                                    m.get_value_for_test_check(
                                        entity_id=entity_id,
                                        check_id=check_id,
                                        monitoring_zone=monitoring_zone,
                                        timestamp=timestamp))
                                   for m in self.metrics])}
                 for monitoring_zone in monitoring_zones])


class MaasStore(object):
    """
    A collection of MaaS configuration objects.
    """
    def __init__(self, clock):
        """
        Initializes the MaaS configuration using the provided clock.

        This initializer reflects the variety of available check types and
        metrics supported by MaaS. Some MaaS check types have been omitted
        for simplicity and clarity. The full list of check types and metrics
        can be found in `the Rackspace Cloud Monitoring Developer Guide, appendix B
            <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/appendix-check-types.html>`_
        """
        self.check_types = {
            'agent.cpu': CheckType(clock=clock, metrics=[
                Metric(name='user_percent_average', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='wait_percent_average', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='sys_percent_average', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='idle_percent_average', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='irq_percent_average', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='usage_average', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='min_cpu_usage', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='max_cpu_usage', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='stolen_percent_average', type=METRIC_TYPE_NUMBER, unit='percent')]),
            'agent.disk': CheckType(clock=clock, metrics=[
                Metric(name='queue', type=METRIC_TYPE_INTEGER),
                Metric(name='read_bytes', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='reads', type=METRIC_TYPE_INTEGER, unit='count'),
                Metric(name='rtime', type=METRIC_TYPE_INTEGER),
                Metric(name='wtime', type=METRIC_TYPE_INTEGER),
                Metric(name='write_bytes', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='writes', type=METRIC_TYPE_INTEGER, unit='count')]),
            'agent.filesystem': CheckType(clock=clock, metrics=[
                Metric(name='avail', type=METRIC_TYPE_INTEGER, unit='kilobytes'),
                Metric(name='free', type=METRIC_TYPE_INTEGER, unit='kilobytes'),
                Metric(name='options', type=METRIC_TYPE_STRING, unit='string'),
                Metric(name='total', type=METRIC_TYPE_INTEGER, unit='kilobytes'),
                Metric(name='used', type=METRIC_TYPE_INTEGER, unit='kilobytes'),
                Metric(name='files', type=METRIC_TYPE_INTEGER, unit='count'),
                Metric(name='free_files', type=METRIC_TYPE_INTEGER, unit='count')]),
            'agent.load_average': CheckType(clock=clock, metrics=[
                Metric(name='1m', type=METRIC_TYPE_NUMBER),
                Metric(name='5m', type=METRIC_TYPE_NUMBER),
                Metric(name='10m', type=METRIC_TYPE_NUMBER)]),
            'agent.memory': CheckType(clock=clock, metrics=[
                Metric(name='actual_free', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='actual_used', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='free', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='ram', type=METRIC_TYPE_INTEGER, unit='megabytes'),
                Metric(name='swap_free', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='swap_page_in', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='swap_page_out', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='swap_total', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='swap_used', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='total', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='used', type=METRIC_TYPE_INTEGER, unit='bytes')]),
            'agent.network': CheckType(clock=clock, metrics=[
                Metric(name='rx_bytes', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='rx_dropped', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='rx_errors', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='rx_packets', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='tx_bytes', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='tx_dropped', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='tx_errors', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='tx_packets', type=METRIC_TYPE_INTEGER, unit='bytes')]),
            'remote.http': CheckType(clock=clock, metrics=[
                Metric(name='bytes', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='cert_end', type=METRIC_TYPE_INTEGER),
                Metric(name='cert_end_in', type=METRIC_TYPE_INTEGER),
                Metric(name='cert_error', type=METRIC_TYPE_STRING, unit='string'),
                Metric(name='cert_issuer', type=METRIC_TYPE_STRING, unit='string'),
                Metric(name='cert_start', type=METRIC_TYPE_INTEGER),
                Metric(name='cert_subject', type=METRIC_TYPE_STRING, unit='string'),
                Metric(name='cert_subject_alternative_names', type=METRIC_TYPE_STRING, unit='string'),
                Metric(name='code', type=METRIC_TYPE_STRING, unit='string'),
                Metric(name='duration', type=METRIC_TYPE_INTEGER),
                Metric(name='truncated', type=METRIC_TYPE_INTEGER, unit='bytes'),
                Metric(name='tt_connect', type=METRIC_TYPE_INTEGER),
                Metric(name='tt_firstbyte', type=METRIC_TYPE_INTEGER)]),
            'remote.ping': CheckType(clock=clock, metrics=[
                Metric(name='available', type=METRIC_TYPE_NUMBER, unit='percent'),
                Metric(name='average', type=METRIC_TYPE_NUMBER),
                Metric(name='count', type=METRIC_TYPE_INTEGER, unit='count'),
                Metric(name='maximum', type=METRIC_TYPE_NUMBER),
                Metric(name='minimum', type=METRIC_TYPE_NUMBER)])}

        self.alarm_states = []

    def latest_alarm_states_for_entity(self, entity_id):
        """
        Computes the latest alarm states for the specified entity.

        Newer alarm states are assumed to be always appended to the list of
        alarm states.
        """
        alarm_states_for_entity = [state for state in self.alarm_states
                                   if state.entity_id == entity_id]
        latest_alarm_states_by_alarm = {}
        for state in alarm_states_for_entity:
            latest_alarm_states_by_alarm[state.alarm_id] = state
        return latest_alarm_states_by_alarm.values()



"""
Canned response for Noit
"""
import time
from mimic.canned_responses.noit_metrics_fixture import (metrics_common_template,
                                                         metrics)


noit_cache = {}


def noit_check_template(request):
    """
    Template used to create check.
    """
    check_template = {
        "check": {
            "attributes": {
                "name": request["name"],
                "module": request["module"],
                "target": request["target"],
                "period": request["period"],
                "timeout": request["timeout"],
                "filterset": request["filterset"]
            },
            "config": {
                "code": ".*",
                "header_123key.test": "400",
                "header_message123": "test",
                "method": "GET",
                "url": "http://127.0.0.1:32321/.well-known/404",
                "redirects": 10,
                "read_limit": 4567777
            },
            "state": {
                "running": "false",
                "killed": "false",
                "configured": "true",
                "disabled": "false",
                "target_ip": "23.253.6.64",
                "last_run": {
                    "@now": "1422323039.361",
                    "#text": "1422323039.357"
                },
                "runtime": "0.958",
                "availability": "available",
                "state": "good",
                "status": "code=200,rt=0.957s,bytes=44779,sslerror",
                "metrics": [
                    {
                        "@type": "inprogress"
                    },
                    {
                        "@type": "current",
                        "@timestamp": str(time.time())
                    }
                ]
            }
        }
    }
    return check_template


def create_check(request, check_id):
    """
    Create the check and saves it in the cache.
    """
    noit_cache[check_id] = noit_check_template(request)
    return noit_cache[check_id]


def test_check(check_type):
    """
    Creates a response for test check
    """
    metrics_common_template["check"]["state"]["metrics"][1].update(metrics.get(check_type, {}))
    return metrics_common_template


def get_check(check_id):
    """
    Gets the check_id from noit_cache
    """
    return noit_cache[check_id]


def get_all_checks():
    """
    Gets all checks from noit_cache
    """
    get_all_check_common_template = {
        "checks": {
            "@resolve_targets": "true",
            "@max_initial_stutter": "30000",
            "@filterset": "default",
            "@transient_min_period": "1000",
            "@transient_period_granularity": "500",
            "@backingstore": "/tmp/noit-bs",
            "config": {
                "@xmlns:ip_acl": "noit://module/ip_acl",
                "ip_acl:whitelist": None
            }
        }
    }
    check_list = []
    for key, value in noit_cache.items():
        each_check = value["check"]["attributes"]
        each_check["uuid"] = key
        each_check["config"] = value["check"]["config"]
        check_list.append(each_check)
    get_all_check_common_template["checks"]["check"] = check_list
    return get_all_check_common_template


def delete_check(check_id):
    """
    Delete the check_id from noit cache
    """
    try:
        del noit_cache[check_id]
    except:
        return 404
    return



# -*- test-case-name: mimic.test.test_customer -*-
"""
API Mock for the Customer API.
"""

import json

from mimic.rest.mimicapp import MimicApp


class CustomerApi(object):
    """
    Rest endpoints for mocked Customer api.
    """

    app = MimicApp()

    def __init__(self, core):
        """
        :param MimicCore core: The core to which this Customer Api will be
        communicating.
        """
        self.core = core

    @app.route('/<string:tenant_id>/contacts', methods=['GET'])
    def get_customer_contacts_for_tenant(self, request, tenant_id):
        """
        Responds with code 200 and returns a list of contacts for the given tenant.

        Until the contacts are set by the `POST` call, this returns the default
        contacts for a tenant.
        """
        response = self.core.contacts_store.list_contacts_for_tenant(tenant_id)
        return json.dumps(response)

    @app.route('/<string:tenant_id>/contacts', methods=['POST'])
    def add_customer_contacts_for_tenant(self, request, tenant_id):
        """
        Adds new contacts to a tenant and responds with a 200.

        Note: If there is a GET on the tenant before this `POST` call, the default
        contacts would have been listed. This POST will overwrite the existing contacts
        and only set the contacts provided.
        """
        content = json.loads(request.content.read())
        contact_list = [(each_contact["email"], each_contact["role"]) for each_contact in content]
        self.core.contacts_store.add_to_contacts_store(tenant_id, contact_list)
        return b''
